# Authors:
#   Thierry Bordaz <tbordaz@redhat.com>
#
# Copyright (C) 2019  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ipalib import Registry, errors
from ipalib import Updater
from ipapython.dn import DN

logger = logging.getLogger(__name__)

register = Registry()


@register()
class update_unhashed_password(Updater):
    """
    DS
    """
    def __remove_update(self, update, key, value):
        statement = dict(action='remove', attr=key, value=value)
        update.setdefault('updates', []).append(statement)

    def __add_update(self, update, key, value):
        statement = dict(action='add', attr=key, value=value)
        update.setdefault('updates', []).append(statement)

    def execute(self, **options):
        logger.debug("Upgrading unhashed password configuration")
        ldap = self.api.Backend.ldap2
        base_config = DN(('cn', 'config'))
        try:
            entry = ldap.get_entry(base_config,
                                   ['nsslapd-unhashed-pw-switch'])
        except errors.NotFound:
            logger.error("Unhashed password configuration not found")
            return False, []

        config_dn = entry.dn

        toggle = entry.single_value.get("nsslapd-unhashed-pw-switch")
        if toggle.lower() not in ['off', 'on', 'nolog']:
            logger.error("Unhashed password had invalid value '%s'", toggle)

        # Check if it exists winsync agreements
        searchfilter = '(objectclass=nsDSWindowsReplicationAgreement)'
        try:
            winsync_agmts, _truncated = ldap.find_entries(
                base_dn=base_config,
                filter=searchfilter,
                attrs_list=[]
            )
        except errors.NotFound:
            logger.debug("Unhashed password this is not a winsync deployment")
            winsync_agmts = []

        update = {
            'dn': config_dn,
            'updates': [],
        }
        if len(winsync_agmts) > 0:
            # We are running in a winsync environment
            # Log a warning that changelog will contain sensitive data
            try:
                cldb_e = ldap.get_entry(
                    DN(('cn', 'changelog5'),
                       ('cn', 'config')),
                    ['nsslapd-changelogdir'])
                cldb = cldb_e.single_value.get("nsslapd-changelogdir")
                logger.warning("This server is configured for winsync, "
                               "the changelog files under %s "
                               "may contain clear text passwords.\n"
                               "Please ensure that these files can be accessed"
                               " only by trusted accounts.\n", cldb)
            except errors.NotFound:
                logger.warning("This server is configured for winsync, "
                               "the changelog files may contain "
                               "clear text passwords.\n"
                               "Please ensure that these files can be accessed"
                               " only by trusted accounts.\n")
            if toggle.lower() == 'on':
                # The current DS configuration already logs the
                # unhashed password
                updates = []
            else:
                self.__remove_update(update, 'nsslapd-unhashed-pw-switch',
                                     toggle)
                self.__add_update(update, 'nsslapd-unhashed-pw-switch', 'on')
                updates = [update]
        else:
            if toggle.lower() == 'nolog':
                updates = []
            else:
                self.__remove_update(update, 'nsslapd-unhashed-pw-switch',
                                     toggle)
                self.__add_update(update, 'nsslapd-unhashed-pw-switch',
                                  'nolog')
                updates = [update]

        return False, updates



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import (
    fuzzy_uuid, fuzzy_automember_message, fuzzy_automember_dn)

from ipatests.test_xmlrpc.tracker.base import Tracker
from ipatests.util import assert_deepequal

from ipalib import api
from ipapython.dn import DN


class AutomemberTracker(Tracker):
    """ Class for tracking automembers """
    retrieve_keys = {u'dn', u'cn', u'member_host', u'description',
                     u'member_automember', u'memberindirect_host',
                     u'automemberinclusiveregex', u'automemberexclusiveregex',
                     u'automembertargetgroup'}
    retrieve_all_keys = retrieve_keys | {u'objectclass',
                                         u'automembertargetgroup'}

    create_keys = retrieve_all_keys
    update_keys = retrieve_keys - {u'dn'}

    add_member_keys = retrieve_keys | {u'member_host'}
    add_condition_keys = retrieve_keys - {u'dn'} |\
        {u'automemberinclusiveregex', u'automembertargetgroup'}
    add_condition_negative_keys = {u'automemberinclusiveregex'}

    def __init__(self, groupname, membertype, description=u'Automember desc'):
        super(AutomemberTracker, self).__init__(default_version=None)
        self.cn = groupname
        self.description = description
        self.membertype = membertype
        self.dn = DN(('cn', self.cn), ('cn', self.membertype.title()),
                     ('cn', 'automember'), ('cn', 'etc'), api.env.basedn)

    def make_create_command(self, *args, **kwargs):
        """ Make function that creates an automember using 'automember-add' """
        return self.make_command('automember_add', self.cn,
                                 description=self.description,
                                 type=self.membertype,
                                 *args, **kwargs)

    def make_delete_command(self):
        """ Make function that deletes an automember using 'automember-del' """
        return self.make_command('automember_del', self.cn,
                                 **dict(type=self.membertype))

    def make_retrieve_command(self, all=False, raw=False, membertype=None):
        """ Make function that retrieves an automember
        using 'automember-show' """
        if membertype is None:
            membertype = self.membertype
        return self.make_command('automember_show', self.cn, type=membertype)

    def make_find_command(self, *args, **kwargs):
        """ Make function that searches for an automember
            using 'automember-find' """
        return self.make_command('automember_find', self.cn,
                                 type=self.membertype)

    def make_update_command(self, updates):
        """ Make function that updates an automember using 'automember-mod' """
        return self.make_command('automember_mod', self.cn,
                                 type=self.membertype, **updates)

    def make_add_member_command(self, options={}):
        """ Make function that adds a member to an automember """
        return self.make_command('automember_add_member', self.cn, **options)

    def make_remove_member_command(self, options={}):
        """ Make function that removes a member from an automember """
        return self.make_command('automember_remove_member',
                                 self.cn, **options)

    def make_rebuild_command(self, *args, **kwargs):
        """ Make function that issues automember_rebuild.
        This function can be executed with arbitrary automember tracker """
        return self.make_command('automember_rebuild', *args, **kwargs)

    def make_add_condition_command(self, *args, **kwargs):
        """ Make function that issues automember_add_condition """
        return self.make_command('automember_add_condition', self.cn,
                                 *args, **kwargs)

    def track_create(self):
        """ Updates expected state for automember creation"""
        self.attrs = dict(
            dn=self.dn,
            mepmanagedentry=[DN(('cn', self.cn), ('cn', 'ng'),
                                ('cn', 'alt'), api.env.basedn)],
            cn=[self.cn],
            description=[self.description],
            ipauniqueid=[fuzzy_uuid],
            objectclass=objectclasses.automember,
            automembertargetgroup=[DN(('cn', self.cn),
                                      ('cn', self.membertype + 's'),
                                      ('cn', 'accounts'), api.env.basedn)]

            )
        self.exists = True

    def add_member(self, options):
        """ Add a member host to automember and perform check """
        if u'group' in options:
            try:
                self.attrs[u'group'] =\
                    self.attrs[u'group'] + [options[u'group']]
            except KeyError:
                self.attrs[u'group'] = [options[u'group']]
            # search for hosts in the target automember and
            # add them as memberindirect hosts
        elif u'hostgroup' in options:
            try:
                self.attrs[u'hostgroup'] =\
                    self.attrs[u'hostgroup'] + [options[u'hostgroup']]
            except KeyError:
                self.attrs[u'hostgroup'] = [options[u'hostgroup']]

        command = self.make_add_member_command(options)
        result = command()
        self.check_add_member(result)

    def remove_member(self, options):
        """ Remove a member host from automember and perform check """
        if u'host' in options:
            self.attrs[u'member_host'].remove(options[u'host'])
        elif u'automember' in options:
            self.attrs[u'member_automember'].remove(options[u'automember'])

        try:
            if not self.attrs[u'member_host']:
                del self.attrs[u'member_host']
        except KeyError:
            pass
        try:
            if not self.attrs[u'member_automember']:
                del self.attrs[u'member_automember']
        except KeyError:
            pass

        command = self.make_remove_member_command(options)
        result = command()
        self.check_remove_member(result)

    def update(self, updates, expected_updates=None):
        """Helper function to update this user and check the result

        Overriding Tracker method for setting self.attrs correctly;
         * most attributes stores its value in list
         * the rest can be overridden by expected_updates
         * allow deleting parametrs if update value is None
        """
        if expected_updates is None:
            expected_updates = {}

        self.ensure_exists()
        command = self.make_update_command(updates)
        result = command()

        for key, value in updates.items():
            if value is None:
                del self.attrs[key]
            else:
                self.attrs[key] = [value]
        for key, value in expected_updates.items():
            if value is None:
                del self.attrs[key]
            else:
                self.attrs[key] = value

        self.check_update(
            result,
            extra_keys=set(updates.keys()) | set(expected_updates.keys())
        )

    def add_condition(self, key, type, inclusiveregex):
        """ Add a condition with given inclusive regex and check for result.
        Only one condition can be added. For more specific uses please
        use make_add_condition_command instead. """
        command = self.make_add_condition_command(
            key=key, type=type, automemberinclusiveregex=inclusiveregex)
        self.attrs['automemberinclusiveregex'] = [u'%s=%s' %
                                                  (key, inclusiveregex[0])]
        result = command()
        self.check_add_condition(result)

    def rebuild(self, no_wait=False):
        """ Rebuild automember conditions and check for result """
        command = self.make_rebuild_command(type=self.membertype,
                                            no_wait=no_wait)
        result = command()
        self.check_rebuild(result, no_wait=no_wait)

    def check_rebuild(self, result, no_wait=False):
        """ Check result of automember_rebuild command """
        if no_wait is False:
            assert_deepequal(dict(
                value=None, result=dict(),
                summary=fuzzy_automember_message
                ), result)
        else:
            assert_deepequal(dict(
                value=None,
                result=dict(dn=fuzzy_automember_dn),
                summary=u'Automember rebuild membership task started'
                ), result)

    def check_add_condition(self, result):
        """ Check result of automember_add_condition command """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Added condition(s) to "%s"' % self.cn,
            completed=1,
            failed=dict(
                failed=dict(automemberinclusiveregex=tuple(),
                            automemberexclusiveregex=tuple(),
                            )
            ),
            result=self.filter_attrs(self.add_condition_keys)
            ), result)

    def check_add_condition_negative(self, result):
        """ Check result of automember_add_condition command
        when the operation didn't add anything. """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Added condition(s) to "%s"' % self.cn,
            completed=0,
            failed=dict(
                failed=dict(automemberinclusiveregex=tuple(),
                            automemberexclusiveregex=tuple(),
                            )
            ),
            result=self.filter_attrs(self.add_condition_negative_keys)
            ), result)

    def check_create(self, result):
        """ Checks 'automember_add' command result """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Added automember rule "%s"' % self.cn,
            result=self.filter_attrs(self.create_keys)
            ), result)

    def check_delete(self, result):
        """ Checks 'automember_del' command result """
        assert_deepequal(dict(
            value=[self.cn],
            summary=u'Deleted automember rule "%s"' % self.cn,
            result=dict(failed=[]),
            ), result)

    def check_retrieve(self, result, all=False, raw=False):
        """ Checks 'automember_show' command result """
        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        assert_deepequal(dict(
            value=self.cn,
            summary=None,
            result=expected
            ), result)

    def check_find(self, result, all=False, raw=False):
        """ Checks 'automember_find' command result """
        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        assert_deepequal(dict(
            count=1,
            truncated=False,
            summary=u'1 rules matched',
            result=[expected],
        ), result)

    def check_update(self, result, extra_keys={}):
        """ Checks 'automember_mod' command result """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Modified automember rule "%s"' % self.cn,
            result=self.filter_attrs(self.update_keys | set(extra_keys))
        ), result)

    def check_add_member(self, result):
        """ Checks 'automember_add_member' command result """
        assert_deepequal(dict(
            completed=1,
            failed={u'member': {u'host': (), u'automember': ()}},
            result=self.filter_attrs(self.add_member_keys)
        ), result)

    def check_add_member_negative(self, result, options):
        """ Checks 'automember_add_member' command result
        when expected result is failure of the operation"""
        expected = dict(
            completed=0,
            failed={u'member': {u'automember': (), u'user': ()}},
            result=self.filter_attrs(self.add_member_keys)
        )
        if u'host' in options:
            expected[u'failed'][u'member'][u'host'] = [(
                options[u'host'], u'no such entry')]
        elif u'automember' in options:
            expected[u'failed'][u'member'][u'automember'] = [(
                options[u'automember'], u'no such entry')]

        assert_deepequal(expected, result)

    def check_remove_member_negative(self, result, options):
        """ Checks 'automember_remove_member' command result
        when expected result is failure of the operation"""
        expected = dict(
            completed=0,
            failed={u'member': {u'automember': (), u'host': ()}},
            result=self.filter_attrs(self.add_member_keys)
        )
        if u'user' in options:
            expected[u'failed'][u'member'][u'host'] = [(
                options[u'user'], u'This entry is not a member')]
        elif u'automember' in options:
            expected[u'failed'][u'member'][u'automember'] = [(
                options[u'automember'], u'This entry is not a member')]

        assert_deepequal(expected, result)

    def check_remove_member(self, result):
        """ Checks 'automember_remove_member' command result """
        self.check_add_member(result)



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import fuzzy_uuid, fuzzy_sudocmddn

from ipatests.test_xmlrpc.tracker.base import Tracker
from ipatests.util import assert_deepequal


class SudoCmdTracker(Tracker):
    """ Class for tracking sudo commands """
    retrieve_keys = {u'dn', u'sudocmd', u'description',
                     u'memberof_sudocmdgroup'}
    retrieve_all_keys = retrieve_keys | {u'ipauniqueid', u'objectclass'}

    create_keys = retrieve_all_keys
    update_keys = retrieve_keys - {u'dn'}

    find_keys = {u'dn', u'sudocmd', u'description'}
    find_all_keys = retrieve_all_keys

    def __init__(self, command, description="Test sudo command"):
        super(SudoCmdTracker, self).__init__(default_version=None)
        self.cmd = command
        self.dn = fuzzy_sudocmddn
        self.description = description

    @property
    def name(self):
        """ Property holding the name of the entry in LDAP """
        return self.cmd

    def make_create_command(self):
        """ Make function that creates a sudocmd using 'sudocmd-add' """
        return self.make_command('sudocmd_add', self.cmd,
                                 description=self.description)

    def make_delete_command(self):
        """ Make function that deletes a sudocmd using 'sudocmd-del' """
        return self.make_command('sudocmd_del', self.cmd)

    def make_retrieve_command(self, all=False, raw=False):
        """ Make function that retrieves a sudocmd using 'sudocmd-show' """
        return self.make_command('sudocmd_show', self.cmd, all=all)

    def make_find_command(self, *args, **kwargs):
        """ Make function that searches for a sudocmd using 'sudocmd-find' """
        return self.make_command('sudocmd_find', *args, **kwargs)

    def make_update_command(self, updates):
        """ Make function that updates a sudocmd using 'sudocmd-mod' """
        return self.make_command('sudocmd_mod', self.cmd, **updates)

    def track_create(self):
        """ Updates expected state for sudocmd creation"""
        self.attrs = dict(
            dn=self.dn,
            sudocmd=[self.cmd],
            description=[self.description],
            ipauniqueid=[fuzzy_uuid],
            objectclass=objectclasses.sudocmd,
            )
        self.exists = True

    def check_create(self, result):
        """ Checks 'sudocmd_add' command result """
        assert_deepequal(dict(
            value=self.cmd,
            summary=u'Added Sudo Command "%s"' % self.cmd,
            result=self.filter_attrs(self.create_keys)
            ), result)

    def check_delete(self, result):
        """ Checks 'sudocmd_del' command result """
        assert_deepequal(dict(
            value=[self.cmd],
            summary=u'Deleted Sudo Command "%s"' % self.cmd,
            result=dict(failed=[]),
            ), result)

    def check_retrieve(self, result, all=False, raw=False):
        """ Checks 'sudocmd_show' command result """
        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        assert_deepequal(dict(
            value=self.cmd,
            summary=None,
            result=expected
            ), result)

    def check_find(self, result, all=False, raw=False):
        """ Checks 'sudocmd_find' command result """
        if all:
            expected = self.filter_attrs(self.find_all_keys)
        else:
            expected = self.filter_attrs(self.find_keys)

        assert_deepequal(dict(
            count=1,
            truncated=False,
            summary=u'1 Sudo Command matched',
            result=[expected],
        ), result)

    def check_update(self, result, extra_keys={}):
        """ Checks 'sudocmd_mod' command result """
        assert_deepequal(dict(
            value=self.cmd,
            summary=u'Modified Sudo Command "%s"' % self.cmd,
            result=self.filter_attrs(self.update_keys | set(extra_keys))
        ), result)



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Trust tests
"""

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
from ipatests.test_webui.task_range import range_tasks
import pytest

ENTITY = 'trust'
CONFIG_ENTITY = 'trustconfig'

CONFIG_DATA = {
    'mod': [
        ['combobox', 'ipantfallbackprimarygroup', 'admins'],
    ]
}

CONFIG_DATA2 = {
    'mod': [
        ['combobox', 'ipantfallbackprimarygroup', 'Default SMB Group']
    ]
}


@pytest.mark.tier1
class trust_tasks(UI_driver):

    def get_data(self, add_data=None):

        domain = self.config.get('ad_domain')

        if not add_data:
            add_data = self.get_add_data()

        data = {
            'pkey': domain,
            'add': add_data,
            'mod': [
                ('multivalued', 'ipantsidblacklistincoming', [
                    ('del', 'S-1-5-18'),
                    ('add', 'S-1-5-21'),
                ]),
                ('multivalued', 'ipantsidblacklistoutgoing', [
                    ('del', 'S-1-5-18'),
                    ('add', 'S-1-5-21'),
                ]),
            ],
        }

        return data

    def get_add_data(self, range_type=None, base_id=None, range_size=None):

        domain = self.config.get('ad_domain')
        admin = self.config.get('ad_admin')
        psw = self.config.get('ad_password')

        add = [
            ('textbox', 'realm_server', domain),
            ('textbox', 'realm_admin', admin),
            ('password', 'realm_passwd', psw),
        ]

        if range_type:
            add.append(('radio', 'range_type', range_type))
        if base_id:
            add.append(('textbox', 'base_id', base_id))
        if range_size:
            add.append(('textbox', 'range_size', range_size))

        return add

    def get_range_name(self):
        domain = self.config.get('ad_domain')
        return domain.upper() + '_id_range'


@pytest.mark.tier1
class test_trust(trust_tasks):

    def setup(self, *args, **kwargs):
        super(test_trust, self).setup(*args, **kwargs)
        if not self.has_trusts():
            self.skip('Trusts not configured')

    @screenshot
    def test_crud(self):
        """
        Basic basic CRUD: trust

        Test establishing trust by using Windows admin credentials
        """
        self.init_app()
        data = self.get_data()
        self.navigate_to_entity('idrange')
        self.delete_record(self.get_range_name())
        self.basic_crud(ENTITY, data)
        self.navigate_to_entity('idrange')
        self.delete_record(self.get_range_name())

    @screenshot
    def test_range_types(self):

        self.init_app()

        r_tasks = range_tasks()
        r_tasks.driver = self.driver
        r_tasks.config = self.config
        r_tasks.get_shifts()
        range_form = r_tasks.get_add_form_data('')
        base_id = range_form.base_id
        range_size = range_form.size
        range_pkey = self.get_range_name()
        column = 'iparangetype'

        self.navigate_to_entity('idrange')
        self.delete_record(range_pkey)

        add = self.get_add_data('ipa-ad-trust', base_id, range_size)
        data = self.get_data(add_data=add)
        self.add_record(ENTITY, data, delete=True)
        self.navigate_to_entity('idrange')
        self.assert_record_value('Active Directory domain range', range_pkey, column)
        self.delete_record(range_pkey)

        self.request_timeout = 60
        add = self.get_add_data('ipa-ad-trust-posix', base_id, range_size)
        data = self.get_data(add_data=add)
        self.add_record(ENTITY, data, delete=True)
        self.navigate_to_entity('idrange')
        self.assert_record_value('Active Directory trust range with POSIX attributes', range_pkey, column)
        self.delete_record(range_pkey)

    @screenshot
    def test_config_mod(self):

        self.init_app()
        self.navigate_to_entity(CONFIG_ENTITY)

        self.mod_record(CONFIG_ENTITY, CONFIG_DATA)
        self.mod_record(CONFIG_ENTITY, CONFIG_DATA2)



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#

"""
Contains Debian-specific service class implementations.
"""

from __future__ import absolute_import

from ipaplatform.base import services as base_services
from ipaplatform.redhat import services as redhat_services
from ipapython import ipautil
from ipaplatform.paths import paths

# Mappings from service names as FreeIPA code references to these services
# to their actual systemd service names
debian_system_units = redhat_services.redhat_system_units.copy()

# For beginning just remap names to add .service
# As more services will migrate to systemd, unit names will deviate and
# mapping will be kept in this dictionary
debian_system_units['httpd'] = 'apache2.service'
debian_system_units['kadmin'] = 'krb5-admin-server.service'
debian_system_units['krb5kdc'] = 'krb5-kdc.service'
debian_system_units['named-regular'] = 'bind9.service'
debian_system_units['named-pkcs11'] = 'bind9-pkcs11.service'
debian_system_units['named'] = debian_system_units['named-pkcs11']
debian_system_units['pki-tomcatd'] = 'pki-tomcatd.service'
debian_system_units['pki_tomcatd'] = debian_system_units['pki-tomcatd']
debian_system_units['ods-enforcerd'] = 'opendnssec-enforcer.service'
debian_system_units['ods_enforcerd'] = debian_system_units['ods-enforcerd']
debian_system_units['ods-signerd'] = 'opendnssec-signer.service'
debian_system_units['ods_signerd'] = debian_system_units['ods-signerd']
debian_system_units['rpcgssd'] = 'rpc-gssd.service'
debian_system_units['rpcidmapd'] = 'nfs-idmapd.service'
debian_system_units['smb'] = 'smbd.service'

# Service classes that implement Debian family-specific behaviour

class DebianService(redhat_services.RedHatService):
    system_units = debian_system_units


class DebianSysvService(base_services.PlatformService):
    def __wait_for_open_ports(self, instance_name=""):
        """
        If this is a service we need to wait for do so.
        """
        ports = None
        if instance_name in base_services.wellknownports:
            ports = base_services.wellknownports[instance_name]
        else:
            if self.service_name in base_services.wellknownports:
                ports = base_services.wellknownports[self.service_name]
        if ports:
            ipautil.wait_for_open_ports('localhost', ports, self.api.env.startup_timeout)

    def stop(self, instance_name='', capture_output=True):
        ipautil.run([paths.SBIN_SERVICE, self.service_name, "stop",
                     instance_name], capture_output=capture_output)
        super(DebianSysvService, self).stop(instance_name)

    def start(self, instance_name='', capture_output=True, wait=True):
        ipautil.run([paths.SBIN_SERVICE, self.service_name, "start",
                     instance_name], capture_output=capture_output)
        if wait and self.is_running(instance_name):
            self.__wait_for_open_ports(instance_name)
        super(DebianSysvService, self).start(instance_name)

    def restart(self, instance_name='', capture_output=True, wait=True):
        ipautil.run([paths.SBIN_SERVICE, self.service_name, "restart",
                     instance_name], capture_output=capture_output)
        if wait and self.is_running(instance_name):
            self.__wait_for_open_ports(instance_name)

    def is_running(self, instance_name="", wait=True):
        ret = True
        try:
            result = ipautil.run([paths.SBIN_SERVICE,
                                  self.service_name, "status",
                                  instance_name],
                                  capture_output=True)
            sout = result.output
            if sout.find("NOT running") >= 0:
                ret = False
            if sout.find("stop") >= 0:
                ret = False
            if sout.find("inactive") >= 0:
                ret = False
        except ipautil.CalledProcessError:
                ret = False
        return ret

    def is_installed(self):
        installed = True
        try:
            ipautil.run([paths.SBIN_SERVICE, self.service_name, "status"])
        except ipautil.CalledProcessError as e:
            if e.returncode == 1:
                # service is not installed or there is other serious issue
                installed = False
        return installed

    @staticmethod
    def is_enabled(instance_name=""):
        # Services are always assumed to be enabled when installed
        return True

    @staticmethod
    def enable():
        return True

    @staticmethod
    def disable():
        return True

    @staticmethod
    def install():
        return True

    @staticmethod
    def remove():
        return True


# For services which have no Debian counterpart
class DebianNoService(base_services.PlatformService):
    @staticmethod
    def start():
        return True

    @staticmethod
    def stop():
        return True

    @staticmethod
    def restart():
        return True

    @staticmethod
    def disable():
        return True


# Function that constructs proper Debian-specific server classes for services
# of specified name

def debian_service_class_factory(name, api=None):
    if name == 'dirsrv':
        return redhat_services.RedHatDirectoryService(name, api)
    if name == 'domainname':
        return DebianNoService(name, api)
    if name == 'ipa':
        return redhat_services.RedHatIPAService(name, api)
    if name in ('pki-tomcatd', 'pki_tomcatd'):
        return redhat_services.RedHatCAService(name, api)
    if name == 'ntpd':
        return DebianSysvService("ntp", api)
    return DebianService(name, api)


# Magicdict containing DebianService instances.

class DebianServices(base_services.KnownServices):
    def __init__(self):
        # pylint: disable=ipa-forbidden-import
        import ipalib  # FixMe: break import cycle
        # pylint: enable=ipa-forbidden-import
        services = dict()
        for s in base_services.wellknownservices:
            services[s] = self.service_class_factory(s, ipalib.api)
        # Call base class constructor. This will lock services to read-only
        super(DebianServices, self).__init__(services)

    @staticmethod
    def service_class_factory(name, api=None):
        return debian_service_class_factory(name, api)

# Objects below are expected to be exported by platform module

timedate_services = base_services.timedate_services
service = debian_service_class_factory
knownservices = DebianServices()



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from ipaclient.frontend import CommandOverride
from ipalib.plugable import Registry

register = Registry()


@register(override=True, no_fail=True)
class passwd(CommandOverride):
    def get_args(self):
        for arg in super(passwd, self).get_args():
            if arg.name == 'current_password':
                arg = arg.clone(sortorder=-1)
            yield arg



#
# Copyright (C) 2018  FreeIPA Contributors.  See COPYING for license
#
from __future__ import absolute_import

import six

from abc import ABCMeta, abstractproperty
from collections import namedtuple
import itertools

from ipatests.util import assert_equal
from ipaserver.install.ipa_replica_install import ReplicaInstall

Keyval = namedtuple('Keyval', ['option', 'value'])


class InstallerTestBase(six.with_metaclass(ABCMeta, object)):
    OPTS_DICT = {}

    # don't allow creating classes with tested_cls unspecified
    @abstractproperty
    def tested_cls(self):
        return None

    def setup_class(self):
        """Initializes the tested class so that it can be used later on
        """
        self.tested_cls.make_parser()
        assert \
            getattr(self.tested_cls, 'option_parser', False), \
            ("Unable to generate option parser for {}"
             .format(self.tested_cls.__name__))

        self._populate_opts_dict()

    @classmethod
    def _populate_opts_dict(cls):
        """Populate the class-owned OPTS_DICT with available options
        """
        if not getattr(cls.tested_cls, 'option_parser', False):
            raise RuntimeError("You need to create the parser of the tested "
                               "class first.")

        # add all options from the option groups
        # pylint: disable=no-member
        for opt_group in cls.tested_cls.option_parser.option_groups:
            for opt in opt_group.option_list:
                cls.OPTS_DICT[opt.dest] = opt._short_opts + opt._long_opts
        # add options outside groups
        for opt in cls.tested_cls.option_parser.option_list:
            cls.OPTS_DICT[opt.dest] = opt._short_opts + opt._long_opts

    def parse_cli_args(self, args):
        """Parses the CLI-like arguments and returns them in python objects

        :param args: A string representing the CLI arguments

        :returns: dictionary and a list of parsed options and arguments
        """
        return self.tested_cls.option_parser.parse_args(args.split())

    def get_installer_instance(self, args):
        """Get instance of the configuring class
        """
        parsed_opts = self.parse_cli_args(args)
        cls = self.tested_cls.get_command_class(*parsed_opts)
        command_instance = cls(*parsed_opts)
        return command_instance.init_configurator()

    def combine_options(self, *args):
        return ' '.join(args)

    def all_option_permutations(self, *key_val):
        """Gets all short-option/long-option permutations

        :param key_val: Keyval tuples specifying the options to grab from
                        OPTS_DICT along with the values to assign to them

        :returns: list or list of lists of all permutations
        """
        if len(key_val) == 0:
            return []
        elif len(key_val) == 1:
            val = key_val[0].value
            return ['{opt} {val}'.format(opt=opt, val=val)
                    for opt in self.OPTS_DICT[key_val[0].option]]

        permutation_lists = [self.OPTS_DICT[k.option] for k in key_val]
        permutation = itertools.product(*permutation_lists)

        ret = []
        for p in permutation:
            opt_vals = []
            for i, kv in enumerate(key_val):
                opt_vals.append(
                    '{opt} {val}'.format(
                        opt=p[i], val=kv.value))
            ret.append(opt_vals)
        return ret


class TestReplicaInstaller(InstallerTestBase):
    tested_cls = ReplicaInstall

    PASSWORD = Keyval("auto_password",
                      "c3ca2246bcf309d1b636581ce429da3522a8aec4")
    ADMIN_PASSWORD = Keyval("admin_password", "milan_je_buh123")
    PRINCIPAL = Keyval("principal", "ubercool_guy")

    def test_password_option_DL1(self):
        # OTP enrollment
        for passwd_opt in self.all_option_permutations(self.PASSWORD):
            ic = self.get_installer_instance(passwd_opt)
            assert_equal(ic.password, self.PASSWORD.value)

        # admin principal enrollment
        for adm_password_opt in (
            self.all_option_permutations(self.ADMIN_PASSWORD)
        ):
            ic = self.get_installer_instance(adm_password_opt)
            assert_equal(ic.password, None)
            assert_equal(ic.admin_password, self.ADMIN_PASSWORD.value)

        # if principal is set, we interpret --password as that principal's
        for passwd_opt, principal_opt in (
            self.all_option_permutations(self.PASSWORD, self.PRINCIPAL)
        ):
            ic = self.get_installer_instance(
                self.combine_options(passwd_opt, principal_opt))
            assert_equal(ic.password, None)
            assert_equal(ic.principal, self.PRINCIPAL.value)
            assert_equal(ic.admin_password, self.PASSWORD.value)

        # if principal is set, we interpret --password as that principal's
        # unless admin-password is also specified, in which case it's once
        # again an OTP
        for adm_password_opt, passwd_opt, principal_opt in (
            self.all_option_permutations(
                self.ADMIN_PASSWORD, self.PASSWORD, self.PRINCIPAL)
        ):
            ic = self.get_installer_instance(
                self.combine_options(
                    adm_password_opt, passwd_opt, principal_opt))
            assert_equal(ic.password, self.PASSWORD.value)
            assert_equal(ic.principal, self.PRINCIPAL.value)
            assert_equal(ic.admin_password, self.ADMIN_PASSWORD.value)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
OTP Tokens

Manage OTP tokens.

IPA supports the use of OTP tokens for multi-factor authentication. This
code enables the management of OTP tokens.

EXAMPLES:

 Add a new token:
   ipa otptoken-add --type=totp --owner=jdoe --desc="My soft token"

 Examine the token:
   ipa otptoken-show a93db710-a31a-4639-8647-f15b2c70b78a

 Change the vendor:
   ipa otptoken-mod a93db710-a31a-4639-8647-f15b2c70b78a --vendor="Red Hat"

 Delete a token:
   ipa otptoken-del a93db710-a31a-4639-8647-f15b2c70b78a
""")

register = Registry()


@register()
class otptoken(Object):
    takes_params = (
        parameters.Str(
            'ipatokenuniqueid',
            primary_key=True,
            label=_(u'Unique ID'),
        ),
        parameters.Str(
            'type',
            required=False,
            label=_(u'Type'),
            doc=_(u'Type of the token'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
            doc=_(u'Token description (informational only)'),
        ),
        parameters.Str(
            'ipatokenowner',
            required=False,
            label=_(u'Owner'),
            doc=_(u'Assigned user of the token (default: self)'),
        ),
        parameters.Str(
            'managedby_user',
            required=False,
            label=_(u'Manager'),
            doc=_(u'Assigned manager of the token (default: self)'),
        ),
        parameters.Bool(
            'ipatokendisabled',
            required=False,
            label=_(u'Disabled'),
            doc=_(u'Mark the token as disabled (default: false)'),
        ),
        parameters.DateTime(
            'ipatokennotbefore',
            required=False,
            label=_(u'Validity start'),
            doc=_(u'First date/time the token can be used'),
        ),
        parameters.DateTime(
            'ipatokennotafter',
            required=False,
            label=_(u'Validity end'),
            doc=_(u'Last date/time the token can be used'),
        ),
        parameters.Str(
            'ipatokenvendor',
            required=False,
            label=_(u'Vendor'),
            doc=_(u'Token vendor name (informational only)'),
        ),
        parameters.Str(
            'ipatokenmodel',
            required=False,
            label=_(u'Model'),
            doc=_(u'Token model (informational only)'),
        ),
        parameters.Str(
            'ipatokenserial',
            required=False,
            label=_(u'Serial'),
            doc=_(u'Token serial (informational only)'),
        ),
        parameters.Bytes(
            'ipatokenotpkey',
            required=False,
            label=_(u'Key'),
            doc=_(u'Token secret (Base32; default: random)'),
        ),
        parameters.Str(
            'ipatokenotpalgorithm',
            required=False,
            label=_(u'Algorithm'),
            doc=_(u'Token hash algorithm'),
        ),
        parameters.Int(
            'ipatokenotpdigits',
            required=False,
            label=_(u'Digits'),
            doc=_(u'Number of digits each token code will have'),
        ),
        parameters.Int(
            'ipatokentotpclockoffset',
            required=False,
            label=_(u'Clock offset'),
            doc=_(u'TOTP token / FreeIPA server time difference'),
        ),
        parameters.Int(
            'ipatokentotptimestep',
            required=False,
            label=_(u'Clock interval'),
            doc=_(u'Length of TOTP token code validity'),
        ),
        parameters.Int(
            'ipatokenhotpcounter',
            required=False,
            label=_(u'Counter'),
            doc=_(u'Initial counter for the HOTP token'),
        ),
    )


@register()
class otptoken_add(Method):
    __doc__ = _("Add a new OTP token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            required=False,
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Str(
            'type',
            required=False,
            cli_metavar="['totp', 'hotp', 'TOTP', 'HOTP']",
            label=_(u'Type'),
            doc=_(u'Type of the token'),
            default=u'totp',
            autofill=True,
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'Token description (informational only)'),
        ),
        parameters.Str(
            'ipatokenowner',
            required=False,
            cli_name='owner',
            label=_(u'Owner'),
            doc=_(u'Assigned user of the token (default: self)'),
        ),
        parameters.Bool(
            'ipatokendisabled',
            required=False,
            cli_name='disabled',
            label=_(u'Disabled'),
            doc=_(u'Mark the token as disabled (default: false)'),
        ),
        parameters.DateTime(
            'ipatokennotbefore',
            required=False,
            cli_name='not_before',
            label=_(u'Validity start'),
            doc=_(u'First date/time the token can be used'),
        ),
        parameters.DateTime(
            'ipatokennotafter',
            required=False,
            cli_name='not_after',
            label=_(u'Validity end'),
            doc=_(u'Last date/time the token can be used'),
        ),
        parameters.Str(
            'ipatokenvendor',
            required=False,
            cli_name='vendor',
            label=_(u'Vendor'),
            doc=_(u'Token vendor name (informational only)'),
        ),
        parameters.Str(
            'ipatokenmodel',
            required=False,
            cli_name='model',
            label=_(u'Model'),
            doc=_(u'Token model (informational only)'),
        ),
        parameters.Str(
            'ipatokenserial',
            required=False,
            cli_name='serial',
            label=_(u'Serial'),
            doc=_(u'Token serial (informational only)'),
        ),
        parameters.Bytes(
            'ipatokenotpkey',
            required=False,
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Token secret (Base32; default: random)'),
            default_from=DefaultFrom(lambda : None),
            # FIXME:
            # lambda: os.urandom(KEY_LENGTH)
            autofill=True,
        ),
        parameters.Str(
            'ipatokenotpalgorithm',
            required=False,
            cli_name='algo',
            cli_metavar="['sha1', 'sha256', 'sha384', 'sha512']",
            label=_(u'Algorithm'),
            doc=_(u'Token hash algorithm'),
            default=u'sha1',
            autofill=True,
        ),
        parameters.Int(
            'ipatokenotpdigits',
            required=False,
            cli_name='digits',
            cli_metavar="['6', '8']",
            label=_(u'Digits'),
            doc=_(u'Number of digits each token code will have'),
            default=6,
            autofill=True,
        ),
        parameters.Int(
            'ipatokentotpclockoffset',
            required=False,
            cli_name='offset',
            label=_(u'Clock offset'),
            doc=_(u'TOTP token / FreeIPA server time difference'),
            default=0,
            autofill=True,
        ),
        parameters.Int(
            'ipatokentotptimestep',
            required=False,
            cli_name='interval',
            label=_(u'Clock interval'),
            doc=_(u'Length of TOTP token code validity'),
            default=30,
            autofill=True,
        ),
        parameters.Int(
            'ipatokenhotpcounter',
            required=False,
            cli_name='counter',
            label=_(u'Counter'),
            doc=_(u'Initial counter for the HOTP token'),
            default=0,
            autofill=True,
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'qrcode',
            required=False,
            label=_(u'(deprecated)'),
            exclude=('cli', 'webui'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_qrcode',
            label=_(u'Do not display QR code'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class otptoken_add_managedby(Method):
    __doc__ = _("Add users that can manage this token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'user',
            required=False,
            multivalue=True,
            cli_name='users',
            label=_(u'member user'),
            doc=_(u'users to add'),
            alwaysask=True,
        ),
    )
    has_output = (
        output.Entry(
            'result',
        ),
        output.Output(
            'failed',
            dict,
            doc=_(u'Members that could not be added'),
        ),
        output.Output(
            'completed',
            int,
            doc=_(u'Number of members added'),
        ),
    )


@register()
class otptoken_del(Method):
    __doc__ = _("Delete an OTP token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            multivalue=True,
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class otptoken_find(Method):
    __doc__ = _("Search for OTP token.")

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'ipatokenuniqueid',
            required=False,
            cli_name='id',
            label=_(u'Unique ID'),
        ),
        parameters.Str(
            'type',
            required=False,
            cli_metavar="['totp', 'hotp', 'TOTP', 'HOTP']",
            label=_(u'Type'),
            doc=_(u'Type of the token'),
            default=u'totp',
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'Token description (informational only)'),
        ),
        parameters.Str(
            'ipatokenowner',
            required=False,
            cli_name='owner',
            label=_(u'Owner'),
            doc=_(u'Assigned user of the token (default: self)'),
        ),
        parameters.Bool(
            'ipatokendisabled',
            required=False,
            cli_name='disabled',
            label=_(u'Disabled'),
            doc=_(u'Mark the token as disabled (default: false)'),
        ),
        parameters.DateTime(
            'ipatokennotbefore',
            required=False,
            cli_name='not_before',
            label=_(u'Validity start'),
            doc=_(u'First date/time the token can be used'),
        ),
        parameters.DateTime(
            'ipatokennotafter',
            required=False,
            cli_name='not_after',
            label=_(u'Validity end'),
            doc=_(u'Last date/time the token can be used'),
        ),
        parameters.Str(
            'ipatokenvendor',
            required=False,
            cli_name='vendor',
            label=_(u'Vendor'),
            doc=_(u'Token vendor name (informational only)'),
        ),
        parameters.Str(
            'ipatokenmodel',
            required=False,
            cli_name='model',
            label=_(u'Model'),
            doc=_(u'Token model (informational only)'),
        ),
        parameters.Str(
            'ipatokenserial',
            required=False,
            cli_name='serial',
            label=_(u'Serial'),
            doc=_(u'Token serial (informational only)'),
        ),
        parameters.Str(
            'ipatokenotpalgorithm',
            required=False,
            cli_name='algo',
            cli_metavar="['sha1', 'sha256', 'sha384', 'sha512']",
            label=_(u'Algorithm'),
            doc=_(u'Token hash algorithm'),
            default=u'sha1',
        ),
        parameters.Int(
            'ipatokenotpdigits',
            required=False,
            cli_name='digits',
            cli_metavar="['6', '8']",
            label=_(u'Digits'),
            doc=_(u'Number of digits each token code will have'),
            default=6,
        ),
        parameters.Int(
            'ipatokentotpclockoffset',
            required=False,
            cli_name='offset',
            label=_(u'Clock offset'),
            doc=_(u'TOTP token / FreeIPA server time difference'),
            default=0,
        ),
        parameters.Int(
            'ipatokentotptimestep',
            required=False,
            cli_name='interval',
            label=_(u'Clock interval'),
            doc=_(u'Length of TOTP token code validity'),
            default=30,
        ),
        parameters.Int(
            'ipatokenhotpcounter',
            required=False,
            cli_name='counter',
            label=_(u'Counter'),
            doc=_(u'Initial counter for the HOTP token'),
            default=0,
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("id")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class otptoken_mod(Method):
    __doc__ = _("Modify a OTP token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'Token description (informational only)'),
        ),
        parameters.Str(
            'ipatokenowner',
            required=False,
            cli_name='owner',
            label=_(u'Owner'),
            doc=_(u'Assigned user of the token (default: self)'),
        ),
        parameters.Bool(
            'ipatokendisabled',
            required=False,
            cli_name='disabled',
            label=_(u'Disabled'),
            doc=_(u'Mark the token as disabled (default: false)'),
        ),
        parameters.DateTime(
            'ipatokennotbefore',
            required=False,
            cli_name='not_before',
            label=_(u'Validity start'),
            doc=_(u'First date/time the token can be used'),
        ),
        parameters.DateTime(
            'ipatokennotafter',
            required=False,
            cli_name='not_after',
            label=_(u'Validity end'),
            doc=_(u'Last date/time the token can be used'),
        ),
        parameters.Str(
            'ipatokenvendor',
            required=False,
            cli_name='vendor',
            label=_(u'Vendor'),
            doc=_(u'Token vendor name (informational only)'),
        ),
        parameters.Str(
            'ipatokenmodel',
            required=False,
            cli_name='model',
            label=_(u'Model'),
            doc=_(u'Token model (informational only)'),
        ),
        parameters.Str(
            'ipatokenserial',
            required=False,
            cli_name='serial',
            label=_(u'Serial'),
            doc=_(u'Token serial (informational only)'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the OTP token object'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class otptoken_remove_managedby(Method):
    __doc__ = _("Remove users that can manage this token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'user',
            required=False,
            multivalue=True,
            cli_name='users',
            label=_(u'member user'),
            doc=_(u'users to remove'),
            alwaysask=True,
        ),
    )
    has_output = (
        output.Entry(
            'result',
        ),
        output.Output(
            'failed',
            dict,
            doc=_(u'Members that could not be removed'),
        ),
        output.Output(
            'completed',
            int,
            doc=_(u'Number of members removed'),
        ),
    )


@register()
class otptoken_show(Method):
    __doc__ = _("Display information about an OTP token.")

    takes_args = (
        parameters.Str(
            'ipatokenuniqueid',
            cli_name='id',
            label=_(u'Unique ID'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

"""
Test certificate requests with IP addresses in SAN, and error
scenarios.

Tests use the default profile (caIPAserviceCert) and the main CA.
The operator is ``admin``.

Various DNS records are created, modified and deleted during this
test.

"""

import ipaddress
import pytest

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from ipalib import api, errors
from ipatests.test_util import yield_fixture
from ipatests.test_xmlrpc.tracker.host_plugin import HostTracker
from ipatests.test_xmlrpc.tracker.user_plugin import UserTracker
from ipatests.test_xmlrpc.xmlrpc_test import XMLRPC_test

host_fqdn = f'iptest.{api.env.domain}'
host_princ = f'host/{host_fqdn}'
host_ptr = f'{host_fqdn}.'

other_fqdn = f'other.{api.env.domain}'
other_ptr = f'{other_fqdn}.'

ipv4_address = '169.254.0.42'
ipv4_revzone_s = '0.254.169.in-addr.arpa.'
ipv4_revrec_s = '42'

ipv6_address = 'fe80::8f18:bdab:4299:95fa'
ipv6_revzone_s = '0.0.0.0.0.0.0.0.0.0.0.0.0.8.e.f.ip6.arpa.'
ipv6_revrec_s = 'a.f.5.9.9.9.2.4.b.a.d.b.8.1.f.8'


@pytest.fixture(scope='class')
def host(request):
    tr = HostTracker('iptest')
    return tr.make_fixture(request)


def _zone_setup(host, zone):
    try:
        host.run_command('dnszone_add', zone)
    except errors.DuplicateEntry:
        delete = False
    else:
        delete = True

    yield zone

    if delete:
        host.run_command('dnszone_del', zone)


def _record_setup(host, zone, record, **kwargs):
    try:
        host.run_command('dnsrecord_add', zone, record, **kwargs)
    except (errors.DuplicateEntry, errors.EmptyModlist):
        delete = False
    else:
        delete = True

    yield

    if delete:
        host.run_command('dnsrecord_del', zone, record, **kwargs)


@yield_fixture(scope='class')
def ipv4_revzone(host):
    yield from _zone_setup(host, ipv4_revzone_s)


@yield_fixture(scope='class')
def ipv6_revzone(host):
    yield from _zone_setup(host, ipv6_revzone_s)


@yield_fixture(scope='class')
def ipv4_ptr(host, ipv4_revzone):
    yield from _record_setup(
        host, ipv4_revzone, ipv4_revrec_s, ptrrecord=host_ptr)


@yield_fixture(scope='class')
def ipv6_ptr(host, ipv6_revzone):
    yield from _record_setup(
        host, ipv6_revzone, ipv6_revrec_s, ptrrecord=host_ptr)


@yield_fixture(scope='class')
def ipv4_a(host):
    yield from _record_setup(
        host, api.env.domain, 'iptest', arecord=ipv4_address)


@yield_fixture(scope='class')
def ipv6_aaaa(host):
    yield from _record_setup(
        host, api.env.domain, 'iptest', aaaarecord=ipv6_address)


@yield_fixture(scope='class')
def other_forward_records(host):
    """
    Create A and AAAA records (to the "correct" IP address) for
    the name "other.{domain}.".

    """
    yield from _record_setup(
        host, api.env.domain, 'other',
        arecord=ipv4_address, aaaarecord=ipv6_address)


@yield_fixture(scope='function')
def ipv4_ptr_other(host, ipv4_revzone):
    yield from _record_setup(
        host, ipv4_revzone, ipv4_revrec_s, ptrrecord=other_ptr)


@yield_fixture(scope='class')
def cname1(host):
    yield from _record_setup(
        host, api.env.domain, 'cname1', cnamerecord='iptest')


@yield_fixture(scope='class')
def cname2(host):
    yield from _record_setup(
        host, api.env.domain, 'cname2', cnamerecord='cname1')


@pytest.fixture(scope='module')
def private_key():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )


def csr(altnames, cn=host_fqdn):
    """
    Return a fixture that generates a CSR with the given altnames.
    The altname values MUST be of type x509.DNSName, x509.IPAddress, etc.

    The subject DN is always to CN={host_fqdn}.

    """
    def inner(private_key):
        builder = x509.CertificateSigningRequestBuilder()
        builder = builder.subject_name(
            x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, cn)]))
        if len(altnames) > 0:
            builder = builder.add_extension(
                x509.SubjectAlternativeName(altnames), False)
        csr = builder.sign(private_key, hashes.SHA256(), default_backend())
        return csr.public_bytes(serialization.Encoding.PEM).decode('ascii')

    return pytest.fixture(scope='module')(inner)


csr_ipv4 = csr([
    x509.DNSName(host_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
])
csr_ipv6 = csr([
    x509.DNSName(host_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv6_address)),
])
csr_ipv4_ipv6 = csr([
    x509.DNSName(host_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
    x509.IPAddress(ipaddress.ip_address(ipv6_address)),
])
csr_extra_ipv4 = csr([
    x509.DNSName(host_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
    x509.IPAddress(ipaddress.ip_address('172.16.254.254')),
])
csr_no_dnsname = csr([
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
])
csr_alice = csr([
    x509.DNSName(host_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
], cn='alice')
csr_iptest_other = csr([
    x509.DNSName(host_fqdn),
    x509.DNSName(other_fqdn),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
])
csr_cname1 = csr([
    x509.DNSName(f'cname1.{api.env.domain}'),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
])
csr_cname2 = csr([
    x509.DNSName(f'cname2.{api.env.domain}'),
    x509.IPAddress(ipaddress.ip_address(ipv4_address)),
])


@pytest.fixture
def user_alice(request):
    user = UserTracker('alice', 'Alice', 'Able')
    return user.make_fixture(request)


# Patterns for ValidationError messages
PAT_FWD = "unreachable from DNS names"
PAT_REV = "does not have PTR record"
PAT_LOOP = "does not match A/AAAA records"
PAT_USER = "forbidden for user principals"


@pytest.mark.tier1
class TestIPAddressSANIssuance(XMLRPC_test):
    """
    These are the tests that can be executed using the "correct" DNS
    records.  Tests for failure scenarios that require "incorrect"
    records are found in other classes.

    """
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(self, host, ipv4_a, ipv4_ptr, csr_ipv4):
        host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(self, host, ipv6_aaaa, ipv6_ptr, csr_ipv6):
        host.run_command('cert_request', csr_ipv6, principal=host_princ)

    def test_issuance_ipv4_ipv6(
            self, host, ipv4_a, ipv6_aaaa, ipv4_ptr, ipv6_ptr, csr_ipv4_ipv6):
        host.run_command('cert_request', csr_ipv4_ipv6, principal=host_princ)

    def test_failure_extra_ip(self, host, ipv4_a, ipv4_ptr, csr_extra_ipv4):
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command(
                'cert_request', csr_extra_ipv4, principal=host_princ)

    def test_failure_no_dnsname(self, host, ipv4_a, ipv4_ptr, csr_no_dnsname):
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command(
                'cert_request', csr_no_dnsname, principal=host_princ)

    def test_failure_user_princ(
            self, host, ipv4_a, ipv4_ptr, csr_alice, user_alice):
        user_alice.ensure_exists()
        with pytest.raises(errors.ValidationError, match=PAT_USER):
            host.run_command(
                'cert_request', csr_alice, principal=user_alice.uid)


@pytest.mark.tier1
class TestIPAddressSANMissingARecord(XMLRPC_test):
    """When there is no A record for the DNS name."""
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(
            self, host, ipv6_aaaa, ipv6_ptr, ipv4_ptr, csr_ipv4):
        """Issuing with IPv4 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(
            self, host, ipv6_aaaa, ipv6_ptr, ipv4_ptr, csr_ipv6):
        """Issuing with IPv6 address succeeds."""
        host.run_command('cert_request', csr_ipv6, principal=host_princ)

    def test_issuance_ipv4_ipv6(
            self, host, ipv6_aaaa, ipv4_ptr, ipv6_ptr, csr_ipv4_ipv6):
        """Issuing with IPv4 *and* IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command(
                'cert_request', csr_ipv4_ipv6, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressSANMissingAAAARecord(XMLRPC_test):
    """When there is no AAAA record for the DNS name."""
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(
            self, host, ipv4_a, ipv6_ptr, ipv4_ptr, csr_ipv4):
        """Issuing with IPv4 address suceeds."""
        host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(
            self, host, ipv4_a, ipv6_ptr, ipv4_ptr, csr_ipv6):
        """Issuing with IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command('cert_request', csr_ipv6, principal=host_princ)

    def test_issuance_ipv4_ipv6(
            self, host, ipv4_a, ipv4_ptr, ipv6_ptr, csr_ipv4_ipv6):
        """Issuing with IPv4 *and* IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command(
                'cert_request', csr_ipv4_ipv6, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressSANMissingIPv4Ptr(XMLRPC_test):
    """When there is no IPv4 PTR record for the address."""
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(
            self, host, ipv4_a, ipv6_aaaa, ipv6_ptr, csr_ipv4):
        """Issuing with IPv4 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_REV):
            host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(
            self, host, ipv4_a, ipv6_aaaa, ipv6_ptr, csr_ipv6):
        """Issuing with IPv6 address succeeds."""
        host.run_command('cert_request', csr_ipv6, principal=host_princ)

    def test_issuance_ipv4_ipv6(
            self, host, ipv4_a, ipv6_aaaa, ipv6_ptr, csr_ipv4_ipv6):
        """Issuing with IPv4 *and* IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_REV):
            host.run_command(
                'cert_request', csr_ipv4_ipv6, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressSANMissingIPv6Ptr(XMLRPC_test):
    """When there is no IPv6 PTR record for the address."""
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(
            self, host, ipv4_a, ipv6_aaaa, ipv4_ptr, csr_ipv4):
        """Issuing with IPv4 address succeeds."""
        host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(
            self, host, ipv4_a, ipv6_aaaa, ipv4_ptr, csr_ipv6):
        """Issuing with IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_REV):
            host.run_command('cert_request', csr_ipv6, principal=host_princ)

    def test_issuance_ipv4_ipv6(
            self, host, ipv4_a, ipv6_aaaa, ipv4_ptr, csr_ipv4_ipv6):
        """Issuing with IPv4 *and* IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_REV):
            host.run_command(
                'cert_request', csr_ipv4_ipv6, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressSANOtherForwardRecords(XMLRPC_test):
    """
    A sanity check that we really are only looking at the
    forward records of interest.  We leave out records for
    the DNS name of interest, but create A and AAAA records
    for a different name in the same zone, which point to the
    IP address of interest.  Issuance must fail.

    """
    def test_host_exists(self, host):
        host.ensure_exists()

    def test_issuance_ipv4(
            self, host, other_forward_records, ipv4_ptr, ipv6_ptr, csr_ipv4):
        """Issuing with IPv4 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command('cert_request', csr_ipv4, principal=host_princ)

    def test_issuance_ipv6(
            self, host, other_forward_records, ipv4_ptr, ipv6_ptr, csr_ipv6):
        """Issuing with IPv6 address fails."""
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command('cert_request', csr_ipv6, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressPTRLoopback(XMLRPC_test):
    """
    A PTR record must point back to the name from which the IP
    address was reached.  Even when the PTR points to a name in the
    SAN, unless the aforementioned condition is satisfied, issuance
    must not proceed.

    ORDER IS IMPORTANT for this test (because the ipv4_ptr fixture
    has *class* scope).

    """
    def test_host_exists(self, host):
        host.ensure_exists()
        host.run_command(
            'host_add_principal', host.fqdn, f'host/other.{api.env.domain}')

    def test_failure(self, host, ipv4_a, ipv4_ptr_other, csr_iptest_other):
        """The A and PTR records are not symmetric."""
        with pytest.raises(errors.ValidationError, match=PAT_LOOP):
            host.run_command(
                'cert_request', csr_iptest_other, principal=host_princ)

    def test_success(self, host, ipv4_a, ipv4_ptr, csr_iptest_other):
        """
        The A and PTR records are symmetric.  This test ensures that the
        presence of an extra DNSName does not interfere with the IP address
        validation.

        """
        host.run_command(
            'cert_request', csr_iptest_other, principal=host_princ)


@pytest.mark.tier1
class TestIPAddressCNAME(XMLRPC_test):
    """
    A single level of CNAME indirection is supported.  PTR must be
    symmetric with the *canonical* name.

    Relevant principal aliases or managedby relationships are
    required by the DNSName validation regime.

    """
    def test_host_exists(self, host, cname1, cname2, ipv4_a, ipv4_ptr):
        # for convenience, this test also establishes the DNS
        # record fixtures, which have class scope
        host.ensure_exists()
        host.run_command(
            'host_add_principal', host.fqdn, f'host/cname1.{api.env.domain}')
        host.run_command(
            'host_add_principal', host.fqdn, f'host/cname2.{api.env.domain}')

    def test_one_level(self, host, csr_cname1):
        host.run_command('cert_request', csr_cname1, principal=host_princ)

    def test_two_levels(self, host, csr_cname2):
        with pytest.raises(errors.ValidationError, match=PAT_FWD):
            host.run_command('cert_request', csr_cname2, principal=host_princ)



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#
"""This covers tests for automemberfeature."""

from __future__ import absolute_import

from ipapython.dn import DN

from ipatests.pytest_ipa.integration import tasks
from ipatests.test_integration.base import IntegrationTest


class TestAutounmembership(IntegrationTest):
    """Tests for autounmembership feature.

    This test class covers the tests for this feature described in
    https://pagure.io/389-ds-base/issue/50077 .
    """
    topology = 'line'
    dn = DN(
        ('cn', 'Auto Membership Plugin'), ('cn', 'plugins'),
        ('cn', 'config')
    )

    def automember_group_add(self, group):
        """Add group/automember entry."""
        cmdlist = [['ipa', 'group-add', '--desc=%s' % group, group],
                   ['ipa', 'automember-add', '--type=group', group],
                   ['ipa', 'automember-add-condition',
                    '--key=destinationindicator', '--type=group',
                    '--inclusive-regex=%s' % group, group]]
        for cmd in cmdlist:
            self.master.run_command(cmd)

    def user_automember_setup(self, group, user, first, last):
        """Add user entry."""
        self.master.run_command(['ipa', 'user-add', '--first=%s' % first,
                                 '--last=%s' % last, user,
                                 '--setattr=destinationindicator=%s' % group])

    def user_automember_mod(self, group, user):
        """Modify user entry."""
        self.master.run_command(['ipa', 'user-mod', user,
                                 '--setattr=destinationindicator=%s' % group])

    def remove_user_automember(self, user, raiseonerr=True):
        """Delete user entry."""
        self.master.run_command(['ipa', 'user-del', user],
                                raiseonerr=raiseonerr)

    def automember_group_cleanup(self, group):
        """Cleanup of automember/group entry."""
        cmdlist = [['ipa', 'automember-del', '--type=group', group],
                   ['ipa', 'group-del', group]]
        for cmd in cmdlist:
            self.master.run_command(cmd, raiseonerr=False)

    def is_user_member_of_group(self, user, group):
        """Check membership of user in the group."""
        result = self.master.run_command(['ipa', 'group-show', group])
        return user in result.stdout_text

    def automember_hostgroup_add(self, hostgroup):
        """Add hostgroup/automember entry."""
        cmdlist = [['ipa', 'hostgroup-add', '--desc=%s' % hostgroup,
                   hostgroup],
                   ['ipa', 'automember-add', '--type=hostgroup', hostgroup],
                   ['ipa', 'automember-add-condition',
                    '--key=nshostlocation', '--type=hostgroup',
                    '--inclusive-regex=%s' % hostgroup, hostgroup]]
        for cmd in cmdlist:
            self.master.run_command(cmd)

    def host_automember_setup(self, hostgroup, host):
        """Add host entry."""
        self.master.run_command(['ipa', 'host-add', host, '--location=%s'
                                 % hostgroup, '--force'])

    def host_automember_mod(self, hostgroup, host):
        """Modify host entry."""
        self.master.run_command(['ipa', 'host-mod', host, '--location=%s'
                                 % hostgroup])

    def remove_host_automember(self, host, raiseonerr=True):
        """Delete host entry."""
        self.master.run_command(['ipa', 'host-del', host],
                                raiseonerr=raiseonerr)

    def automember_hostgroup_cleanup(self, hostgroup):
        """Cleanup of automember/hostgroup entry."""
        cmdlist = [['ipa', 'automember-del', '--type=hostgroup', hostgroup],
                   ['ipa', 'hostgroup-del', hostgroup]]
        for cmd in cmdlist:
            self.master.run_command(cmd, raiseonerr=False)

    def is_host_member_of_hostgroup(self, host, hostgroup):
        """Check membership of host in the hostgroup."""
        result = self.master.run_command(['ipa', 'hostgroup-show', hostgroup])
        return host in result.stdout_text

    def check_autounmembership_in_ldap(self, state='on'):
        """Check autounmembership state."""
        conn = self.master.ldap_connect()
        entry = conn.get_entry(self.dn)  # pylint: disable=no-member
        assert state == entry.single_value['automemberprocessmodifyops']

    def disable_autounmembership_in_ldap(self):
        """Disable autounmembership."""
        conn = self.master.ldap_connect()
        entry = conn.get_entry(self.dn)  # pylint: disable=no-member
        entry.single_value['automemberprocessmodifyops'] = 'off'
        conn.update_entry(entry)  # pylint: disable=no-member
        self.master.run_command(['ipactl', 'restart'])

    def test_modify_user_entry_unmembership(self):
        """Test when autounmembership feature is enabled.

        Test that an update on a user entry triggers a re-evaluation of the
        user's automember groups when auto-unmembership is active.
        """
        try:
            # testcase setup
            tasks.kinit_admin(self.master)
            self.check_autounmembership_in_ldap(state='on')
            group1 = 'DallasUsers'
            group2 = 'HoustonUsers'
            user1 = 'user1'
            first = 'first'
            last = 'user'
            self.automember_group_add(group1)
            self.automember_group_add(group2)
            self.user_automember_setup(group1, user1, first, last)
            assert self.is_user_member_of_group(user1, group1)

            # Modifying the user group so that it becomes part of new group
            self.user_automember_mod(group2, user1)
            assert self.is_user_member_of_group(user1, group2)
            assert not self.is_user_member_of_group(user1, group1)

            # Deleting the user entry and checking its not part of group now
            self.remove_user_automember(user1)
            assert not self.is_user_member_of_group(user1, group2)

        finally:
            # testcase cleanup
            self.remove_user_automember(user1, raiseonerr=False)
            self.automember_group_cleanup(group1)
            self.automember_group_cleanup(group2)

    def test_modify_host_entry_unmembership(self):
        """Test when autounmembership feature is enabled.

        Test that an update on a host entry triggers a re-evaluation of the
        host's automember groups when auto-unmembership is active.
        """
        try:
            # testcase setup
            self.check_autounmembership_in_ldap(state='on')
            hostgroup1 = 'Brno'
            hostgroup2 = 'Boston'
            host1 = 'host1.example.com'
            self.automember_hostgroup_add(hostgroup1)
            self.automember_hostgroup_add(hostgroup2)
            self.host_automember_setup(hostgroup1, host1)
            assert self.is_host_member_of_hostgroup(host1, hostgroup1)

            # Modifying the user group so that it becomes part of new group
            self.host_automember_mod(hostgroup2, host1)
            assert self.is_host_member_of_hostgroup(host1, hostgroup2)
            assert not self.is_host_member_of_hostgroup(host1, hostgroup1)

            # Deleting the user entry and checking its not part of group now
            self.remove_host_automember(host1)
            assert not self.is_host_member_of_hostgroup(host1, hostgroup2)

        finally:
            # testcase cleanup
            self.remove_host_automember(host1, raiseonerr=False)
            self.automember_hostgroup_cleanup(hostgroup1)
            self.automember_hostgroup_cleanup(hostgroup2)

    def test_modify_user_entry_unmembership_disabled(self):
        """Test when autounmembership feature is disabled.

        Test that an update on a user entry does not triggers re-evaluation of
        the user's automember groups when auto-unmembership is disabled.
        """
        try:
            # testcase setup
            self.disable_autounmembership_in_ldap()
            self.check_autounmembership_in_ldap(state='off')
            group1 = 'PuneUsers'
            group2 = 'BrnoUsers'
            user2 = 'user2'
            first = 'second'
            last = 'user'
            self.automember_group_add(group1)
            self.automember_group_add(group2)
            self.user_automember_setup(group1, user2, first, last)
            assert self.is_user_member_of_group(user2, group1)

            # Modifying the user group so that it becomes part of new group
            self.user_automember_mod(group2, user2)
            assert not self.is_user_member_of_group(user2, group2)
            assert self.is_user_member_of_group(user2, group1)

            # Running automember-build so that user is part of correct group
            self.master.run_command(['ipa', 'automember-rebuild',
                                     '--users=%s' % user2])
            assert self.is_user_member_of_group(user2, group2)
            assert not self.is_user_member_of_group(user2, group1)

        finally:
            # testcase cleanup
            self.remove_user_automember(user2, raiseonerr=False)
            self.automember_group_cleanup(group1)
            self.automember_group_cleanup(group2)

    def test_modify_host_entry_unmembership_disabled(self):
        """Test when autounmembership feature is disabled.

        Test that an update on a host entry does not triggers re-evaluation of
        the host's automember groups when auto-unmembership is disabled.
        """
        try:
            # testcase setup
            self.check_autounmembership_in_ldap(state='off')
            hostgroup1 = 'Pune'
            hostgroup2 = 'Raleigh'
            host2 = 'host2.example.com'
            self.automember_hostgroup_add(hostgroup1)
            self.automember_hostgroup_add(hostgroup2)
            self.host_automember_setup(hostgroup1, host2)
            assert self.is_host_member_of_hostgroup(host2, hostgroup1)

            # Modifying the user group so that it becomes part of new group
            self.host_automember_mod(hostgroup2, host2)
            assert not self.is_host_member_of_hostgroup(host2, hostgroup2)
            assert self.is_host_member_of_hostgroup(host2, hostgroup1)

            # Running the automember-build so host is part of correct hostgroup
            self.master.run_command(
                ['ipa', 'automember-rebuild', '--hosts=%s' % host2]
            )
            assert self.is_host_member_of_hostgroup(host2, hostgroup2)
            assert not self.is_host_member_of_hostgroup(host2, hostgroup1)

        finally:
            # testcase cleanup
            self.remove_host_automember(host2, raiseonerr=False)
            self.automember_hostgroup_cleanup(hostgroup1)
            self.automember_hostgroup_cleanup(hostgroup2)



# Authors:
#   Petr Viktorin <pviktori@redhat.com>
#
# Copyright (C) 2014  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provides command-line options for very limited Nose compatibility"""

import os
import sys
import logging

from ipapython.ipa_log_manager import Formatter, convert_log_level


def pytest_addoption(parser):
    group = parser.getgroup("IPA nosetests compatibility shim")

    group.addoption('--with-xunit', action="store_const",
           dest="xmlpath", metavar="path",  default=None,
           const=os.environ.get('IPATEST_XUNIT_PATH', './nosetests.xml'),
           help="create junit-xml style report file at $IPATEST_XUNIT_PATH,"
                "or nosetests.xml by default")

    group.addoption('--logging-level', action="store",
           dest="logging_level", metavar="level", default='CRITICAL',
           help="level for logging to stderr. "
                "Bypasses pytest logging redirection."
                "May be used to show progress of long-running tests.")


def pytest_configure(config):
    if config.getoption('logging_level'):
        # Forward IPA logging to a normal Python logger. Nose's logcapture plugin
        # can't work with IPA-managed loggers
        class LogHandler(logging.Handler):
            name = 'forwarding log handler'
            logger = logging.getLogger('IPA')

            def emit(self, record):
                capture = config.pluginmanager.getplugin('capturemanager')
                orig_stdout, orig_stderr = sys.stdout, sys.stderr
                if capture:
                    # pylint: disable=no-member
                    if hasattr(capture, 'suspend_global_capture'):
                        # pytest >= 3.3
                        capture.suspend_global_capture()
                    else:
                        # legacy support for pytest <= 3.2 (Fedora 27)
                        capture._capturing.suspend_capturing()
                    # pylint: enable=no-member
                sys.stderr.write(self.format(record))
                sys.stderr.write('\n')
                if capture:
                    # pylint: disable=no-member
                    if hasattr(capture, 'resume_global_capture'):
                        capture.resume_global_capture()
                    else:
                        capture._capturing.resume_capturing()
                    # pylint: enable=no-member
                sys.stdout, sys.stderr = orig_stdout, orig_stderr

        level = convert_log_level(config.getoption('logging_level'))

        handler = LogHandler()
        handler.setFormatter(Formatter('[%(name)s] %(message)s'))
        handler.setLevel(level)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
Set a user's password

If someone other than a user changes that user's password (e.g., Helpdesk
resets it) then the password will need to be changed the first time it
is used. This is so the end-user is the only one who knows the password.

The IPA password policy controls how often a password may be changed,
what strength requirements exist, and the length of the password history.

EXAMPLES:

 To reset your own password:
   ipa passwd

 To change another user's password:
   ipa passwd tuser1
""")

register = Registry()


@register()
class passwd(Command):
    __doc__ = _("Set a user's password.")

    takes_args = (
        parameters.Str(
            'principal',
            cli_name='user',
            label=_(u'User name'),
            default_from=DefaultFrom(lambda : None),
            # FIXME:
            # lambda: util.get_current_principal()
            autofill=True,
            no_convert=True,
        ),
        parameters.Password(
            'password',
            label=_(u'New Password'),
            confirm=True,
        ),
        parameters.Password(
            'current_password',
            label=_(u'Current Password'),
            default_from=DefaultFrom(lambda principal: None, 'principal'),
            # FIXME:
            # lambda principal: get_current_password(principal)
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            bool,
            doc=_(u'True means the operation was successful'),
        ),
        output.Output(
            'value',
            unicode,
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#   Pavel Zuna <pzuna@redhat.com>
#   John Dennis <jdennis@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Test the `ipaserver/plugins/role.py` module.
"""

from ipalib import api, errors
from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import Declarative, fuzzy_uuid
from ipapython.dn import DN
import pytest

search = u'test-role'

role1 = u'test-role-1'
role1_dn = DN(('cn',role1),api.env.container_rolegroup,
              api.env.basedn)
renamedrole1 = u'test-role'
invalidrole1 = u' whitespace '

role2 = u'test-role-2'
role2_dn = DN(('cn',role2),api.env.container_rolegroup,
              api.env.basedn)

group1 = u'testgroup1'
group1_dn = DN(('cn',group1),api.env.container_group,
               api.env.basedn)

privilege1 = u'r,w privilege 1'
privilege1_dn = DN(('cn', privilege1), DN(api.env.container_privilege),
                   api.env.basedn)


@pytest.mark.tier1
class test_role(Declarative):

    cleanup_commands = [
        ('role_del', [role1], {}),
        ('role_del', [role2], {}),
        ('group_del', [group1], {}),
        ('privilege_del', [privilege1], {}),
    ]

    tests = [

        dict(
            desc='Try to retrieve non-existent %r' % role1,
            command=('role_show', [role1], {}),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Try to update non-existent %r' % role1,
            command=('role_mod', [role1], dict(description=u'Foo')),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Try to delete non-existent %r' % role1,
            command=('role_del', [role1], {}),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Try to rename non-existent %r' % role1,
            command=('role_mod', [role1], dict(setattr=u'cn=%s' % renamedrole1)),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Search for non-existent %r' % role1,
            command=('role_find', [role1], {}),
            expected=dict(
                count=0,
                truncated=False,
                summary=u'0 roles matched',
                result=[],
            ),
        ),


        dict(
            desc='Create invalid %r' % invalidrole1,
            command=('role_add', [invalidrole1],
                dict(description=u'role desc 1')
            ),
            expected=errors.ValidationError(name='name',
                error=u'Leading and trailing spaces are not allowed'),
        ),


        dict(
            desc='Create %r' % role1,
            command=('role_add', [role1],
                dict(description=u'role desc 1')
            ),
            expected=dict(
                value=role1,
                summary=u'Added role "%s"' % role1,
                result=dict(
                    dn=role1_dn,
                    cn=[role1],
                    description=[u'role desc 1'],
                    objectclass=objectclasses.role,
                ),
            ),
        ),


        dict(
            desc='Retrieve %r' % role1,
            command=('role_show', [role1], {}),
            expected=dict(
                value=role1,
                summary=None,
                result=dict(
                    dn=role1_dn,
                    cn=[role1],
                    description=[u'role desc 1'],
                ),
            ),
        ),


        dict(
            desc='Create %r' % group1,
            command=(
                'group_add', [group1], dict(description=u'group desc 1',
                nonposix=True,)
            ),
            expected=dict(
                value=group1,
                summary=u'Added group "testgroup1"',
                result=dict(
                    dn=group1_dn,
                    cn=[group1],
                    description=[u'group desc 1'],
                    objectclass=objectclasses.group,
                    ipauniqueid=[fuzzy_uuid],
                ),
            ),
        ),


        dict(
            desc='Create %r' % privilege1,
            command=('privilege_add', [privilege1],
                dict(description=u'privilege desc. 1')
            ),
            expected=dict(
                value=privilege1,
                summary=u'Added privilege "%s"' % privilege1,
                result=dict(
                    dn=privilege1_dn,
                    cn=[privilege1],
                    description=[u'privilege desc. 1'],
                    objectclass=objectclasses.privilege,
                ),
            ),
        ),


        dict(
            desc='Add privilege %r to role %r' % (privilege1, role1),
            command=('role_add_privilege', [role1],
                dict(privilege=privilege1)
            ),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        privilege=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'role desc 1'],
                    'memberof_privilege': [privilege1],
                }
            ),
        ),


        dict(
            desc='Add zero privileges to role %r' % role1,
            command=('role_add_privilege', [role1], dict(privilege=None)
            ),
            expected=dict(
                completed=0,
                failed=dict(
                    member=dict(
                        privilege=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'role desc 1'],
                    'memberof_privilege': [privilege1],
                }
            ),
        ),


        dict(
            desc='Remove zero privileges from role %r' % role1,
            command=('role_remove_privilege', [role1], dict(privilege=None)
            ),
            expected=dict(
                completed=0,
                failed=dict(
                    member=dict(
                        privilege=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'role desc 1'],
                    'memberof_privilege': [privilege1],
                }
            ),
        ),


        dict(
            desc='Add member %r to %r' % (group1, role1),
            command=('role_add_member', [role1], dict(group=group1)),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        user=[],
                        group=[],
                        host=[],
                        hostgroup=[],
                        service=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'role desc 1'],
                    'member_group': [group1],
                    'memberof_privilege': [privilege1],
                }
            ),
        ),


        dict(
            desc='Retrieve %r to verify member-add' % role1,
            command=('role_show', [role1], {}),
            expected=dict(
                value=role1,
                summary=None,
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'role desc 1'],
                    'member_group': [group1],
                    'memberof_privilege': [privilege1],
                },
            ),
        ),


        dict(
            desc='Search for %r with members' % role1,
            command=('role_find', [role1], {'no_members': False}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                        'member_group': [group1],
                        'memberof_privilege': [privilege1],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r' % role1,
            command=('role_find', [role1], {}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r with members' % search,
            command=('role_find', [search], {'no_members': False}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                        'member_group': [group1],
                        'memberof_privilege': [privilege1],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r' % search,
            command=('role_find', [search], {}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                    },
                ],
            ),
        ),


        dict(
            desc='Create %r' % role2,
            command=('role_add', [role2],
                dict(description=u'role desc 2')
            ),
            expected=dict(
                value=role2,
                summary=u'Added role "%s"' % role2,
                result=dict(
                    dn=role2_dn,
                    cn=[role2],
                    description=[u'role desc 2'],
                    objectclass=objectclasses.role,
                ),
            ),
        ),


        dict(
            desc='Search for %r with members' % role1,
            command=('role_find', [role1], {'no_members': False}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                        'member_group': [group1],
                        'memberof_privilege': [privilege1],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r' % role1,
            command=('role_find', [role1], {}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r with members' % search,
            command=('role_find', [search], {'no_members': False}),
            expected=dict(
                count=2,
                truncated=False,
                summary=u'2 roles matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                        'member_group': [group1],
                        'memberof_privilege': [privilege1],
                    },
                    {
                        'dn': role2_dn,
                        'cn': [role2],
                        'description': [u'role desc 2'],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r' % search,
            command=('role_find', [search], {}),
            expected=dict(
                count=2,
                truncated=False,
                summary=u'2 roles matched',
                result=[
                    {
                        'dn': role1_dn,
                        'cn': [role1],
                        'description': [u'role desc 1'],
                    },
                    {
                        'dn': role2_dn,
                        'cn': [role2],
                        'description': [u'role desc 2'],
                    },
                ],
            ),
        ),


        dict(
            desc='Update %r' % role1,
            command=(
                'role_mod', [role1], dict(description=u'New desc 1')
            ),
            expected=dict(
                value=role1,
                summary=u'Modified role "%s"' % role1,
                result=dict(
                    cn=[role1],
                    description=[u'New desc 1'],
                    member_group=[group1],
                    memberof_privilege=[privilege1],
                ),
            ),
        ),


        dict(
            desc='Retrieve %r to verify update' % role1,
            command=('role_show', [role1], {}),
            expected=dict(
                value=role1,
                summary=None,
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'New desc 1'],
                    'member_group': [group1],
                    'memberof_privilege': [privilege1],
                },
            ),
        ),


        dict(
            desc='Remove member %r from %r' % (group1, role1),
            command=('role_remove_member', [role1], dict(group=group1)),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        user=[],
                        group=[],
                        host=[],
                        hostgroup=[],
                        service=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'New desc 1'],
                    'memberof_privilege': [privilege1],
                },
            ),
        ),


        dict(
            desc='Retrieve %r to verify member-del' % role1,
            command=('role_show', [role1], {}),
            expected=dict(
                value=role1,
                summary=None,
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'New desc 1'],
                    'memberof_privilege': [privilege1],
                },
            ),
        ),


        dict(
            desc='Delete %r' % group1,
            command=('group_del', [group1], {}),
            expected=dict(
                result=dict(failed=[]),
                value=[group1],
                summary=u'Deleted group "testgroup1"',
            )
        ),


        dict(
            desc='Rename %r' % role1,
            command=('role_mod', [role1], dict(setattr=u'cn=%s' % renamedrole1)),
            expected=dict(
                value=role1,
                result=dict(
                    cn=[renamedrole1],
                    description=[u'New desc 1'],
                    memberof_privilege=[privilege1],
                ),
                summary=u'Modified role "%s"' % role1
            )
        ),


        dict(
            desc='Rename %r back' % renamedrole1,
            command=('role_mod', [renamedrole1], dict(setattr=u'cn=%s' % role1)),
            expected=dict(
                value=renamedrole1,
                result=dict(
                    cn=[role1],
                    description=[u'New desc 1'],
                    memberof_privilege=[privilege1],
                ),
                summary=u'Modified role "%s"' % renamedrole1
            )
        ),


        dict(
            desc='Remove privilege %r from role %r' % (privilege1, role1),
            command=('role_remove_privilege', [role1],
                dict(privilege=privilege1)
            ),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        privilege=[],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'New desc 1'],
                }
            ),
        ),


        dict(
            desc='Remove privilege %r from role %r again' % (privilege1, role1),
            command=('role_remove_privilege', [role1],
                dict(privilege=privilege1)
            ),
            expected=dict(
                completed=0,
                failed=dict(
                    member=dict(
                        privilege=[(u'%s' % privilege1, u'This entry is not a member'),],
                    ),
                ),
                result={
                    'dn': role1_dn,
                    'cn': [role1],
                    'description': [u'New desc 1'],
                }
            ),
        ),



        dict(
            desc='Delete %r' % role1,
            command=('role_del', [role1], {}),
            expected=dict(
                result=dict(failed=[]),
                value=[role1],
                summary=u'Deleted role "%s"' % role1,
            )
        ),


        dict(
            desc='Try to delete non-existent %r' % role1,
            command=('role_del', [role1], {}),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Try to show non-existent %r' % role1,
            command=('role_show', [role1], {}),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Try to update non-existent %r' % role1,
            command=('role_mod', [role1], dict(description=u'Foo')),
            expected=errors.NotFound(reason=u'%s: role not found' % role1),
        ),


        dict(
            desc='Search for %r' % search,
            command=('role_find', [search], {}),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 role matched',
                result=[
                    {
                        'dn': role2_dn,
                        'cn': [role2],
                        'description': [u'role desc 2'],
                    },
                ],
            ),
        ),


        dict(
            desc='Delete %r' % role2,
            command=('role_del', [role2], {}),
            expected=dict(
                result=dict(failed=[]),
                value=[role2],
                summary=u'Deleted role "%s"' % role2,
            )
        ),


        dict(
            desc='Search for %r' % search,
            command=('role_find', [search], {}),
            expected=dict(
                count=0,
                truncated=False,
                summary=u'0 roles matched',
                result=[],
            ),
        ),

    ]



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#   Rob Crittenden <rcritten@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
RPC client and shared RPC client/server functionality.

This module adds some additional functionality on top of the ``xmlrpc.client``
module in the Python standard library (``xmlrpclib`` in Python 2).
For documentation on the ``xmlrpclib`` module, see:

    http://docs.python.org/2/library/xmlrpclib.html

Also see the `ipaserver.rpcserver` module.
"""

from __future__ import absolute_import

from decimal import Decimal
import datetime
import logging
import os
import locale
import base64
import json
import re
import socket
import gzip
import urllib
from ssl import SSLError

from cryptography import x509 as crypto_x509
import gssapi
from dns.exception import DNSException
import six

from ipalib.backend import Connectible
from ipalib.constants import LDAP_GENERALIZED_TIME_FORMAT
from ipalib.errors import (public_errors, UnknownError, NetworkError,
                           XMLRPCMarshallError, JSONError)
from ipalib import errors, capabilities
from ipalib.request import context, Connection
from ipalib.x509 import Encoding as x509_Encoding
from ipapython import ipautil
from ipapython import session_storage
from ipapython.cookie import Cookie
from ipapython.dnsutil import DNSName, query_srv
from ipalib.text import _
from ipalib.util import create_https_connection
from ipalib.krb_utils import KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN, KRB5KRB_AP_ERR_TKT_EXPIRED, \
                             KRB5_FCC_PERM, KRB5_FCC_NOFILE, KRB5_CC_FORMAT, \
                             KRB5_REALM_CANT_RESOLVE, KRB5_CC_NOTFOUND, get_principal
from ipapython.dn import DN
from ipapython.kerberos import Principal
from ipalib.capabilities import VERSION_WITHOUT_CAPABILITIES
from ipalib import api

# The XMLRPC client is in  "six.moves.xmlrpc_client", but pylint
# cannot handle that
try:
    from xmlrpclib import (Binary, Fault, DateTime, dumps, loads, ServerProxy,
            Transport, ProtocolError, MININT, MAXINT)
except ImportError:
    # pylint: disable=import-error
    from xmlrpc.client import (Binary, Fault, DateTime, dumps, loads, ServerProxy,
            Transport, ProtocolError, MININT, MAXINT)

# pylint: disable=import-error
if six.PY3:
    from http.client import RemoteDisconnected
else:
    from httplib import BadStatusLine as RemoteDisconnected
# pylint: enable=import-error


if six.PY3:
    unicode = str

logger = logging.getLogger(__name__)

COOKIE_NAME = 'ipa_session'
CCACHE_COOKIE_KEY = 'X-IPA-Session-Cookie'

errors_by_code = dict((e.errno, e) for e in public_errors)


def update_persistent_client_session_data(principal, data):
    '''
    Given a principal create or update the session data for that
    principal in the persistent secure storage.

    Raises ValueError if unable to perform the action for any reason.
    '''

    try:
        session_storage.store_data(principal, CCACHE_COOKIE_KEY, data)
    except Exception as e:
        raise ValueError(str(e))

def read_persistent_client_session_data(principal):
    '''
    Given a principal return the stored session data for that
    principal from the persistent secure storage.

    Raises ValueError if unable to perform the action for any reason.
    '''

    try:
        return session_storage.get_data(principal, CCACHE_COOKIE_KEY)
    except Exception as e:
        raise ValueError(str(e))

def delete_persistent_client_session_data(principal):
    '''
    Given a principal remove the session data for that
    principal from the persistent secure storage.

    Raises ValueError if unable to perform the action for any reason.
    '''

    try:
        session_storage.remove_data(principal, CCACHE_COOKIE_KEY)
    except Exception as e:
        raise ValueError(str(e))

def xml_wrap(value, version):
    """
    Wrap all ``str`` in ``xmlrpc.client.Binary``.

    Because ``xmlrpc.client.dumps()`` will itself convert all ``unicode`` instances
    into UTF-8 encoded ``str`` instances, we don't do it here.

    So in total, when encoding data for an XML-RPC packet, the following
    transformations occur:

        * All ``str`` instances are treated as binary data and are wrapped in
          an ``xmlrpc.client.Binary()`` instance.

        * Only ``unicode`` instances are treated as character data. They get
          converted to UTF-8 encoded ``str`` instances (although as mentioned,
          not by this function).

    Also see `xml_unwrap()`.

    :param value: The simple scalar or simple compound value to wrap.
    """
    if type(value) in (list, tuple):
        return tuple(xml_wrap(v, version) for v in value)
    if isinstance(value, dict):
        return dict(
            (k, xml_wrap(v, version)) for (k, v) in value.items()
        )
    if type(value) is bytes:
        return Binary(value)
    if type(value) is Decimal:
        # transfer Decimal as a string
        return unicode(value)
    if isinstance(value, int) and (value < MININT or value > MAXINT):
        return unicode(value)
    if isinstance(value, DN):
        return str(value)

    # Encode datetime.datetime objects as xmlrpc.client.DateTime objects
    if isinstance(value, datetime.datetime):
        if capabilities.client_has_capability(version, 'datetime_values'):
            return DateTime(value)
        else:
            return value.strftime(LDAP_GENERALIZED_TIME_FORMAT)

    if isinstance(value, DNSName):
        if capabilities.client_has_capability(version, 'dns_name_values'):
            return {'__dns_name__': unicode(value)}
        else:
            return unicode(value)

    if isinstance(value, Principal):
        return unicode(value)

    if isinstance(value, crypto_x509.Certificate):
        return base64.b64encode(
            value.public_bytes(x509_Encoding.DER)).decode('ascii')

    if isinstance(value, crypto_x509.CertificateSigningRequest):
        return base64.b64encode(
            value.public_bytes(x509_Encoding.DER)).decode('ascii')

    assert type(value) in (unicode, float, int, bool, type(None))
    return value


def xml_unwrap(value, encoding='UTF-8'):
    """
    Unwrap all ``xmlrpc.Binary``, decode all ``str`` into ``unicode``.

    When decoding data from an XML-RPC packet, the following transformations
    occur:

        * The binary payloads of all ``xmlrpc.client.Binary`` instances are
          returned as ``str`` instances.

        * All ``str`` instances are treated as UTF-8 encoded Unicode strings.
          They are decoded and the resulting ``unicode`` instance is returned.

    Also see `xml_wrap()`.

    :param value: The value to unwrap.
    :param encoding: The Unicode encoding to use (defaults to ``'UTF-8'``).
    """
    if isinstance(value, (unicode, int, float, bool)):
        # most common first
        return value
    elif value is None:
        return value
    elif isinstance(value, bytes):
        return value.decode(encoding)
    elif isinstance(value, (list, tuple)):
        return tuple(xml_unwrap(v, encoding) for v in value)
    elif isinstance(value, dict):
        if '__dns_name__' in value:
            return DNSName(value['__dns_name__'])
        else:
            return dict(
                (k, xml_unwrap(v, encoding)) for (k, v) in value.items()
            )
    elif isinstance(value, Binary):
        assert type(value.data) is bytes
        return value.data
    elif isinstance(value, DateTime):
        # xmlprc DateTime is converted to string of %Y%m%dT%H:%M:%S format
        return datetime.datetime.strptime(str(value), "%Y%m%dT%H:%M:%S")
    raise TypeError(value)


def xml_dumps(params, version, methodname=None, methodresponse=False,
              encoding='UTF-8'):
    """
    Encode an XML-RPC data packet, transparently wraping ``params``.

    This function will wrap ``params`` using `xml_wrap()` and will
    then encode the XML-RPC data packet using ``xmlrpc.client.dumps()`` (from the
    Python standard library).

    For documentation on the ``xmlrpc.client.dumps()`` function, see:

        http://docs.python.org/library/xmlrpc.client.html#convenience-functions

    Also see `xml_loads()`.

    :param params: A ``tuple`` or an ``xmlrpc.client.Fault`` instance.
    :param methodname: The name of the method to call if this is a request.
    :param methodresponse: Set this to ``True`` if this is a response.
    :param encoding: The Unicode encoding to use (defaults to ``'UTF-8'``).
    """
    if type(params) is tuple:
        params = xml_wrap(params, version)
    else:
        assert isinstance(params, Fault)
    return dumps(params,
        methodname=methodname,
        methodresponse=methodresponse,
        encoding=encoding,
        allow_none=True,
    )


class _JSONPrimer(dict):
    """Fast JSON primer and pre-converter

    Prepare a data structure for JSON serialization. In an ideal world, priming
    could be handled by the default hook of json.dumps(). Unfortunately the
    hook treats Python 2 str as text while FreeIPA considers str as bytes.

    The primer uses a couple of tricks to archive maximum performance:

    * O(1) type look instead of O(n) chain of costly isinstance() calls
    * __missing__ and __mro__ with caching to handle subclasses
    * inline code with minor code duplication (func lookup in enc_list/dict)
    * avoid surplus function calls (e.g. func is _identity, obj.__class__
      instead if type(obj))
    * function default arguments to turn global into local lookups
    * avoid re-creation of bound method objects (e.g. result.append)
    * on-demand lookup of client capabilities with cached values

    Depending on the client version number, the primer converts:

    * bytes -> {'__base64__': b64encode}
    * datetime -> {'__datetime__': LDAP_GENERALIZED_TIME}
    * DNSName -> {'__dns_name__': unicode}

    The _ipa_obj_hook() functions unserializes the marked JSON objects to
    bytes, datetime and DNSName.

    :see: _ipa_obj_hook
    """
    __slots__ = ('version', '_cap_datetime', '_cap_dnsname')

    _identity = object()

    def __init__(self, version, _identity=_identity):
        super(_JSONPrimer, self).__init__()
        self.version = version
        self._cap_datetime = None
        self._cap_dnsname = None
        self.update({
            unicode: _identity,
            bool: _identity,
            int: _identity,
            type(None): _identity,
            float: _identity,
            Decimal: unicode,
            DN: str,
            Principal: unicode,
            DNSName: self._enc_dnsname,
            datetime.datetime: self._enc_datetime,
            bytes: self._enc_bytes,
            list: self._enc_list,
            tuple: self._enc_list,
            dict: self._enc_dict,
            crypto_x509.Certificate: self._enc_certificate,
            crypto_x509.CertificateSigningRequest: self._enc_certificate,
        })

    def __missing__(self, typ):
        # walk MRO to find best match
        for c in typ.__mro__:
            if c in self:
                self[typ] = self[c]
                return self[c]
        # use issubclass to check for registered ABCs
        for c in self:
            if issubclass(typ, c):
                self[typ] = self[c]
                return self[c]
        raise TypeError(typ)

    def convert(self, obj, _identity=_identity):
        # obj.__class__ is twice as fast as type(obj)
        func = self[obj.__class__]
        return obj if func is _identity else func(obj)

    def _enc_datetime(self, val):
        cap = self._cap_datetime
        if cap is None:
            cap = capabilities.client_has_capability(self.version,
                                                     'datetime_values')
            self._cap_datetime = cap
        if cap:
            return {'__datetime__': val.strftime(LDAP_GENERALIZED_TIME_FORMAT)}
        else:
            return val.strftime(LDAP_GENERALIZED_TIME_FORMAT)

    def _enc_dnsname(self, val):
        cap = self._cap_dnsname
        if cap is None:
            cap = capabilities.client_has_capability(self.version,
                                                     'dns_name_values')
            self._cap_dnsname = cap
        if cap:
            return {'__dns_name__': unicode(val)}
        else:
            return unicode(val)

    def _enc_bytes(self, val):
        encoded = base64.b64encode(val)
        if not six.PY2:
            encoded = encoded.decode('ascii')
        return {'__base64__': encoded}

    def _enc_list(self, val, _identity=_identity):
        result = []
        append = result.append
        for v in val:
            func = self[v.__class__]
            append(v if func is _identity else func(v))
        return result

    def _enc_dict(self, val, _identity=_identity, _iteritems=six.iteritems):
        result = {}
        for k, v in _iteritems(val):
            func = self[v.__class__]
            result[k] = v if func is _identity else func(v)
        return result

    def _enc_certificate(self, val):
        return self._enc_bytes(val.public_bytes(x509_Encoding.DER))


def json_encode_binary(val, version, pretty_print=False):
    """Serialize a Python object structure to JSON

    :param object val: Python object structure
    :param str version: client version
    :param bool pretty_print: indent and sort JSON (warning: slow!)
    :return: text
    :note: pretty printing triggers a slow path in Python's JSON module. Only
           use pretty_print in debug mode.
    """
    result = _JSONPrimer(version).convert(val)
    if pretty_print:
        return json.dumps(result, indent=4, sort_keys=True)
    else:
        return json.dumps(result)


def _ipa_obj_hook(dct, _iteritems=six.iteritems, _list=list):
    """JSON object hook

    :see: _JSONPrimer
    """
    if '__base64__' in dct:
        return base64.b64decode(dct['__base64__'])
    elif '__datetime__' in dct:
        return datetime.datetime.strptime(dct['__datetime__'],
                                          LDAP_GENERALIZED_TIME_FORMAT)
    elif '__dns_name__' in dct:
        return DNSName(dct['__dns_name__'])
    else:
        # XXX tests assume tuples. Is this really necessary?
        for k, v in _iteritems(dct):
            if v.__class__ is _list:
                dct[k] = tuple(v)
        return dct


def json_decode_binary(val):
    """Convert serialized JSON string back to Python data structure

    :param val: JSON string
    :type val: str, bytes
    :return: Python data structure
    :see: _ipa_obj_hook, _JSONPrimer
    """
    if isinstance(val, bytes):
        val = val.decode('utf-8')

    return json.loads(val, object_hook=_ipa_obj_hook)


def decode_fault(e, encoding='UTF-8'):
    assert isinstance(e, Fault)
    if isinstance(e.faultString, bytes):
        return Fault(e.faultCode, e.faultString.decode(encoding))
    return e


def xml_loads(data, encoding='UTF-8'):
    """
    Decode the XML-RPC packet in ``data``, transparently unwrapping its params.

    This function will decode the XML-RPC packet in ``data`` using
    ``xmlrpc.client.loads()`` (from the Python standard library).  If ``data``
    contains a fault, ``xmlrpc.client.loads()`` will itself raise an
    ``xmlrpc.client.Fault`` exception.

    Assuming an exception is not raised, this function will then unwrap the
    params in ``data`` using `xml_unwrap()`.  Finally, a
    ``(params, methodname)`` tuple is returned containing the unwrapped params
    and the name of the method being called.  If the packet contains no method
    name, ``methodname`` will be ``None``.

    For documentation on the ``xmlrpc.client.loads()`` function, see:

        http://docs.python.org/2/library/xmlrpclib.html#convenience-functions

    Also see `xml_dumps()`.

    :param data: The XML-RPC packet to decode.
    """
    try:
        (params, method) = loads(data)
        return (xml_unwrap(params), method)
    except Fault as e:
        raise decode_fault(e)


class DummyParser:
    def __init__(self):
        self.data = []

    def feed(self, data):
        self.data.append(data)

    def close(self):
        return b''.join(self.data)


class MultiProtocolTransport(Transport):
    """Transport that handles both XML-RPC and JSON"""
    def __init__(self, *args, **kwargs):
        Transport.__init__(self)
        self.protocol = kwargs.get('protocol', None)

    def getparser(self):
        if self.protocol == 'json':
            parser = DummyParser()
            return parser, parser
        else:
            return Transport.getparser(self)

    def send_content(self, connection, request_body):
        if self.protocol == 'json':
            connection.putheader("Content-Type", "application/json")
        else:
            connection.putheader("Content-Type", "text/xml")

        # gzip compression would be set up here, but we have it turned off
        # (encode_threshold is None)

        connection.putheader("Content-Length", str(len(request_body)))
        connection.endheaders(request_body)


class LanguageAwareTransport(MultiProtocolTransport):
    """Transport sending Accept-Language header"""
    def get_host_info(self, host):
        host, extra_headers, x509 = MultiProtocolTransport.get_host_info(
            self, host)

        try:
            lang = locale.setlocale(
                locale.LC_MESSAGES, ''
            ).split('.')[0].lower()
        except locale.Error:
            # fallback to default locale
            lang = 'en_us'

        if not isinstance(extra_headers, list):
            extra_headers = []

        extra_headers.append(
            ('Accept-Language', lang.replace('_', '-'))
        )
        extra_headers.append(
            ('Referer', 'https://%s/ipa/xml' % str(host))
        )

        return (host, extra_headers, x509)


class SSLTransport(LanguageAwareTransport):
    """Handles an HTTPS transaction to an XML-RPC server."""
    def make_connection(self, host):
        host, self._extra_headers, _x509 = self.get_host_info(host)

        if self._connection and host == self._connection[0]:
            logger.debug("HTTP connection keep-alive (%s)", host)
            return self._connection[1]

        conn = create_https_connection(
            host, 443,
            getattr(context, 'ca_certfile', None),
            tls_version_min=api.env.tls_version_min,
            tls_version_max=api.env.tls_version_max)

        conn.connect()
        logger.debug("New HTTP connection (%s)", host)

        self._connection = host, conn
        return self._connection[1]


class KerbTransport(SSLTransport):
    """
    Handles Kerberos Negotiation authentication to an XML-RPC server.
    """
    flags = [gssapi.RequirementFlag.mutual_authentication,
             gssapi.RequirementFlag.out_of_sequence_detection]

    def __init__(self, *args, **kwargs):
        SSLTransport.__init__(self, *args, **kwargs)
        self._sec_context = None
        self.service = kwargs.pop("service", "HTTP")
        self.ccache = kwargs.pop("ccache", None)

    def _handle_exception(self, e, service=None):
        minor = e.min_code
        if minor == KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN:
            raise errors.ServiceError(service=service)
        elif minor == KRB5_FCC_NOFILE:
            raise errors.NoCCacheError()
        elif minor == KRB5KRB_AP_ERR_TKT_EXPIRED:
            raise errors.TicketExpired()
        elif minor == KRB5_FCC_PERM:
            raise errors.BadCCachePerms()
        elif minor == KRB5_CC_FORMAT:
            raise errors.BadCCacheFormat()
        elif minor == KRB5_REALM_CANT_RESOLVE:
            raise errors.CannotResolveKDC()
        elif minor == KRB5_CC_NOTFOUND:
            raise errors.CCacheError()
        else:
            raise errors.KerberosError(message=unicode(e))

    def _get_host(self):
        return self._connection[0]

    def _remove_extra_header(self, name):
        for (h, v) in self._extra_headers:
            if h == name:
                self._extra_headers.remove((h, v))
                break

    def get_auth_info(self, use_cookie=True):
        """
        Two things can happen here. If we have a session we will add
        a cookie for that. If not we will set an Authorization header.
        """
        if not isinstance(self._extra_headers, list):
            self._extra_headers = []

        # Remove any existing Cookie first
        self._remove_extra_header('Cookie')
        if use_cookie:
            session_cookie = getattr(context, 'session_cookie', None)
            if session_cookie:
                self._extra_headers.append(('Cookie', session_cookie))
                return

        # Set the remote host principal
        host = self._get_host()
        service = self.service + "@" + host.split(':')[0]

        try:
            creds = None
            if self.ccache:
                creds = gssapi.Credentials(usage='initiate',
                                           store={'ccache': self.ccache})
            name = gssapi.Name(service, gssapi.NameType.hostbased_service)
            self._sec_context = gssapi.SecurityContext(creds=creds, name=name,
                                                       flags=self.flags)
            response = self._sec_context.step()
        except gssapi.exceptions.GSSError as e:
            self._handle_exception(e, service=service)

        self._set_auth_header(response)

    def _set_auth_header(self, token):
        # Remove any existing authorization header first
        self._remove_extra_header('Authorization')

        if token:
            self._extra_headers.append(
                ('Authorization', 'negotiate %s' % base64.b64encode(token).decode('ascii'))
            )

    def _auth_complete(self, response):
        if self._sec_context:
            header = response.getheader('www-authenticate', '')
            token = None
            for field in header.split(','):
                k, _dummy, v = field.strip().partition(' ')
                if k.lower() == 'negotiate':
                    try:
                        token = base64.b64decode(v.encode('ascii'))
                        break
                    # b64decode raises TypeError on invalid input
                    except (TypeError, UnicodeError):
                        pass
            if not token:
                raise errors.KerberosError(
                    message=u"No valid Negotiate header in server response")
            token = self._sec_context.step(token=token)
            if self._sec_context.complete:
                self._sec_context = None
                return True
            self._set_auth_header(token)
            return False
        elif response.status == 401:
            self.get_auth_info(use_cookie=False)
            return False
        return True

    def single_request(self, host, handler, request_body, verbose=0):
        # Based on Python 2.7's xmllib.Transport.single_request
        try:
            h = self.make_connection(host)

            if verbose:
                h.set_debuglevel(1)

            self.get_auth_info()

            while True:
                if six.PY2:
                    # pylint: disable=no-value-for-parameter
                    self.send_request(h, handler, request_body)
                    # pylint: enable=no-value-for-parameter
                    self.send_host(h, host)
                    self.send_user_agent(h)
                    self.send_content(h, request_body)
                    response = h.getresponse(buffering=True)
                else:
                    self.__send_request(h, host, handler, request_body, verbose)
                    response = h.getresponse()

                if response.status != 200:
                    # Must read response (even if it is empty)
                    # before sending another request.
                    #
                    # https://docs.python.org/3/library/http.client.html
                    #   #http.client.HTTPConnection.getresponse
                    #
                    # https://pagure.io/freeipa/issue/7752
                    #
                    response.read()

                    if response.status == 401:
                        if not self._auth_complete(response):
                            continue

                    raise ProtocolError(
                        host + handler,
                        response.status, response.reason,
                        response.msg)

                self.verbose = verbose
                if not self._auth_complete(response):
                    continue
                return self.parse_response(response)
        except gssapi.exceptions.GSSError as e:
            self._handle_exception(e)
        except RemoteDisconnected:
            # keep-alive connection was terminated by remote peer, close
            # connection and let transport handle reconnect for us.
            self.close()
            logger.debug("HTTP server has closed connection (%s)", host)
            raise
        except BaseException as e:
            # Unexpected exception may leave connections in a bad state.
            self.close()
            logger.debug("HTTP connection destroyed (%s)",
                         host, exc_info=True)
            raise

    if six.PY3:
        def __send_request(self, connection, host, handler, request_body, debug):
            # Based on xmlrpc.client.Transport.send_request
            headers = self._extra_headers[:]
            if debug:
                connection.set_debuglevel(1)
            if self.accept_gzip_encoding and gzip:
                connection.putrequest("POST", handler, skip_accept_encoding=True)
                connection.putheader("Accept-Encoding", "gzip")
                headers.append(("Accept-Encoding", "gzip"))
            else:
                connection.putrequest("POST", handler)
            headers.append(("User-Agent", self.user_agent))
            self.send_headers(connection, headers)  # pylint: disable=E1101
            self.send_content(connection, request_body)
            return connection

    # Find all occurrences of the expiry component
    expiry_re = re.compile(r'.*?(&expiry=\d+).*?')

    def _slice_session_cookie(self, session_cookie):
        # Keep only the cookie value and strip away all other info.
        # This is to reduce the churn on FILE ccaches which grow every time we
        # set new data. The expiration time for the cookie is set in the
        # encrypted data anyway and will be enforced by the server
        http_cookie = session_cookie.http_cookie()
        # We also remove the "expiry" part from the data which is not required
        for exp in self.expiry_re.findall(http_cookie):
            http_cookie = http_cookie.replace(exp, '')
        return http_cookie

    def store_session_cookie(self, cookie_header):
        '''
        Given the contents of a Set-Cookie header scan the header and
        extract each cookie contained within until the session cookie
        is located. Examine the session cookie if the domain and path
        are specified, if not update the cookie with those values from
        the request URL. Then write the session cookie into the key
        store for the principal. If the cookie header is None or the
        session cookie is not present in the header no action is
        taken.

        Context Dependencies:

        The per thread context is expected to contain:
            principal
                The current pricipal the HTTP request was issued for.
            request_url
                The URL of the HTTP request.

        '''

        if cookie_header is None:
            return

        principal = getattr(context, 'principal', None)
        request_url = getattr(context, 'request_url', None)
        logger.debug("received Set-Cookie (%s)'%s'", type(cookie_header),
                     cookie_header)

        if not isinstance(cookie_header, list):
            cookie_header = [cookie_header]

        # Search for the session cookie
        session_cookie = None
        try:
            for cookie in cookie_header:
                session_cookie = (
                    Cookie.get_named_cookie_from_string(
                        cookie, COOKIE_NAME, request_url,
                        timestamp=datetime.datetime.utcnow())
                    )
                if session_cookie is not None:
                    break
        except Exception as e:
            logger.error("unable to parse cookie header '%s': %s",
                         cookie_header, e)
            return

        if session_cookie is None:
            return

        cookie_string = self._slice_session_cookie(session_cookie)
        logger.debug("storing cookie '%s' for principal %s",
                     cookie_string, principal)
        try:
            update_persistent_client_session_data(principal, cookie_string)
        except Exception as e:
            # Not fatal, we just can't use the session cookie we were sent.
            pass

    def parse_response(self, response):
        if six.PY2:
            header = response.msg.getheaders('Set-Cookie')
        else:
            header = response.msg.get_all('Set-Cookie')
        self.store_session_cookie(header)
        return SSLTransport.parse_response(self, response)


class DelegatedKerbTransport(KerbTransport):
    """
    Handles Kerberos Negotiation authentication and TGT delegation to an
    XML-RPC server.
    """
    flags = [gssapi.RequirementFlag.delegate_to_peer,
             gssapi.RequirementFlag.mutual_authentication,
             gssapi.RequirementFlag.out_of_sequence_detection]


class RPCClient(Connectible):
    """
    Forwarding backend plugin for XML-RPC client.

    Also see the `ipaserver.rpcserver.xmlserver` plugin.
    """

    # Values to set on subclasses:
    session_path = None
    server_proxy_class = ServerProxy
    protocol = None
    env_rpc_uri_key = None

    def get_url_list(self, rpc_uri):
        """
        Create a list of urls consisting of the available IPA servers.
        """
        # the configured URL defines what we use for the discovered servers
        (_scheme, _netloc, path, _params, _query, _fragment
            ) = urllib.parse.urlparse(rpc_uri)
        servers = []
        name = '_ldap._tcp.%s.' % self.env.domain

        try:
            answers = query_srv(name)
        except DNSException:
            answers = []

        for answer in answers:
            server = str(answer.target).rstrip(".")
            servers.append('https://%s%s' % (ipautil.format_netloc(server), path))

        # make sure the configured master server is there just once and
        # it is the first one.
        if rpc_uri in servers:
            servers.remove(rpc_uri)
        servers.insert(0, rpc_uri)

        return servers

    def get_session_cookie_from_persistent_storage(self, principal):
        '''
        Retrieves the session cookie for the given principal from the
        persistent secure storage. Returns None if not found or unable
        to retrieve the session cookie for any reason, otherwise
        returns a Cookie object containing the session cookie.
        '''

        # Get the session data, it should contain a cookie string
        # (possibly with more than one cookie).
        try:
            cookie_string = read_persistent_client_session_data(principal)
            if cookie_string is None:
                return None
            cookie_string = cookie_string.decode('utf-8')
        except Exception as e:
            logger.debug('Error reading client session data: %s', e)
            return None

        # Search for the session cookie within the cookie string
        try:
            session_cookie = Cookie.get_named_cookie_from_string(
                cookie_string, COOKIE_NAME,
                timestamp=datetime.datetime.utcnow())
        except Exception as e:
            logger.debug(
                'Error retrieving cookie from the persistent storage: %s',
                e)
            return None

        return session_cookie

    def apply_session_cookie(self, url):
        '''
        Attempt to load a session cookie for the current principal
        from the persistent secure storage. If the cookie is
        successfully loaded adjust the input url's to point to the
        session path and insert the session cookie into the per thread
        context for later insertion into the HTTP request. If the
        cookie is not successfully loaded then the original url is
        returned and the per thread context is not modified.

        Context Dependencies:

        The per thread context is expected to contain:
            principal
                The current pricipal the HTTP request was issued for.

        The per thread context will be updated with:
            session_cookie
                A cookie string to be inserted into the Cookie header
                of the HTPP request.

        '''

        original_url = url
        principal = getattr(context, 'principal', None)

        session_cookie = self.get_session_cookie_from_persistent_storage(principal)
        if session_cookie is None:
            logger.debug("failed to find session_cookie in persistent storage "
                         "for principal '%s'",
                         principal)
            return original_url
        else:
            logger.debug("found session_cookie in persistent storage for "
                         "principal '%s', cookie: '%s'",
                         principal, session_cookie)

        # Decide if we should send the cookie to the server
        try:
            session_cookie.http_return_ok(original_url)
        except Cookie.Expired as e:
            logger.debug("deleting session data for principal '%s': %s",
                         principal, e)
            try:
                delete_persistent_client_session_data(principal)
            except Exception as e:
                pass
            return original_url
        except Cookie.URLMismatch as e:
            logger.debug("not sending session cookie, URL mismatch: %s", e)
            return original_url
        except Exception as e:
            logger.error("not sending session cookie, unknown error: %s", e)
            return original_url

        # O.K. session_cookie is valid to be returned, stash it away where it will will
        # get included in a HTTP Cookie headed sent to the server.
        logger.debug("setting session_cookie into context '%s'",
                     session_cookie.http_cookie())
        setattr(context, 'session_cookie', session_cookie.http_cookie())

        # Form the session URL by substituting the session path into the original URL
        scheme, netloc, path, params, query, fragment = urllib.parse.urlparse(original_url)
        path = self.session_path
        # urlencode *can* take one argument
        # pylint: disable=too-many-function-args
        session_url = urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))

        return session_url

    def create_connection(self, ccache=None, verbose=None, fallback=None,
                          delegate=None, ca_certfile=None):
        if verbose is None:
            verbose = self.api.env.verbose
        if fallback is None:
            fallback = self.api.env.fallback
        if delegate is None:
            delegate = self.api.env.delegate
        if ca_certfile is None:
            ca_certfile = self.api.env.tls_ca_cert
        context.ca_certfile = ca_certfile

        rpc_uri = self.env[self.env_rpc_uri_key]
        try:
            principal = get_principal(ccache_name=ccache)
            stored_principal = getattr(context, 'principal', None)
            if principal != stored_principal:
                try:
                    delattr(context, 'session_cookie')
                except AttributeError:
                    pass
            setattr(context, 'principal', principal)
            # We have a session cookie, try using the session URI to see if it
            # is still valid
            if not delegate:
                rpc_uri = self.apply_session_cookie(rpc_uri)
        except (errors.CCacheError, ValueError):
            # No session key, do full Kerberos auth
            pass
        urls = self.get_url_list(rpc_uri)

        proxy_kw = {
            'allow_none': True,
            'encoding': 'UTF-8',
            'verbose': verbose
        }

        for url in urls:
            # should we get ProtocolError (=> error in HTTP response) and
            # 401 (=> Unauthorized), we'll be re-trying with new session
            # cookies several times
            for _try_num in range(0, 5):
                if url.startswith('https://'):
                    if delegate:
                        transport_class = DelegatedKerbTransport
                    else:
                        transport_class = KerbTransport
                else:
                    transport_class = LanguageAwareTransport
                proxy_kw['transport'] = transport_class(
                    protocol=self.protocol, service='HTTP', ccache=ccache)
                logger.debug('trying %s', url)
                setattr(context, 'request_url', url)
                serverproxy = self.server_proxy_class(url, **proxy_kw)
                if len(urls) == 1:
                    # if we have only 1 server and then let the
                    # main requester handle any errors. This also means it
                    # must handle a 401 but we save a ping.
                    return serverproxy
                try:
                    command = getattr(serverproxy, 'ping')
                    try:
                        command([], {})
                    except Fault as e:
                        e = decode_fault(e)
                        if e.faultCode in errors_by_code:
                            error = errors_by_code[e.faultCode]
                            raise error(message=e.faultString)
                        else:
                            raise UnknownError(
                                code=e.faultCode,
                                error=e.faultString,
                                server=url,
                            )
                    # We don't care about the response, just that we got one
                    return serverproxy
                # pylint: disable=try-except-raise
                except errors.KerberosError:
                    # kerberos error on one server is likely on all
                    raise
                # pylint: enable=try-except-raise
                except ProtocolError as e:
                    if hasattr(context, 'session_cookie') and e.errcode == 401:
                        # Unauthorized. Remove the session and try again.
                        delattr(context, 'session_cookie')
                        try:
                            delete_persistent_client_session_data(principal)
                        except Exception:
                            # This shouldn't happen if we have a session but
                            # it isn't fatal.
                            pass
                        # try the same url once more with a new session cookie
                        continue
                    if not fallback:
                        raise
                    else:
                        logger.info(
                            'Connection to %s failed with %s', url, e)
                    # try the next url
                    break
                except Exception as e:
                    if not fallback:
                        raise
                    else:
                        logger.info(
                            'Connection to %s failed with %s', url, e)
                    # try the next url
                    break
        # finished all tries but no serverproxy was found
        raise NetworkError(uri=_('any of the configured servers'),
                           error=', '.join(urls))

    def destroy_connection(self):
        conn = getattr(context, self.id, None)
        if conn is not None:
            conn = conn.conn._ServerProxy__transport
            conn.close()

    def _call_command(self, command, params):
        """Call the command with given params"""
        # For XML, this method will wrap/unwrap binary values
        # For JSON we do that in the proxy
        return command(*params)

    def forward(self, name, *args, **kw):
        """
        Forward call to command named ``name`` over XML-RPC.

        This method will encode and forward an XML-RPC request, and will then
        decode and return the corresponding XML-RPC response.

        :param command: The name of the command being forwarded.
        :param args: Positional arguments to pass to remote command.
        :param kw: Keyword arguments to pass to remote command.
        """
        server = getattr(context, 'request_url', None)
        command = getattr(self.conn, name)
        params = [args, kw]

        # we'll be trying to connect multiple times with a new session cookie
        # each time should we be getting UNAUTHORIZED error from the server
        max_tries = 5
        for try_num in range(0, max_tries):
            logger.debug("[try %d]: Forwarding '%s' to %s server '%s'",
                         try_num + 1, name, self.protocol, server)
            try:
                return self._call_command(command, params)
            except Fault as e:
                e = decode_fault(e)
                logger.debug('Caught fault %d from server %s: %s', e.faultCode,
                             server, e.faultString)
                if e.faultCode in errors_by_code:
                    error = errors_by_code[e.faultCode]
                    raise error(message=e.faultString)
                raise UnknownError(
                    code=e.faultCode,
                    error=e.faultString,
                    server=server,
                )
            except ProtocolError as e:
                # By catching a 401 here we can detect the case where we have
                # a single IPA server and the session is invalid. Otherwise
                # we always have to do a ping().
                session_cookie = getattr(context, 'session_cookie', None)
                if session_cookie and e.errcode == 401:
                    # Unauthorized. Remove the session and try again.
                    delattr(context, 'session_cookie')
                    try:
                        principal = getattr(context, 'principal', None)
                        delete_persistent_client_session_data(principal)
                    except Exception as e:
                        # This shouldn't happen if we have a session
                        # but it isn't fatal.
                        logger.debug("Error trying to remove persisent "
                                     "session data: %s", e)

                    # Create a new serverproxy with the non-session URI
                    serverproxy = self.create_connection(
                        os.environ.get('KRB5CCNAME'), self.env.verbose,
                        self.env.fallback, self.env.delegate)

                    setattr(context, self.id,
                            Connection(serverproxy, self.disconnect))
                    # try to connect again with the new session cookie
                    continue
                raise NetworkError(uri=server, error=e.errmsg)
            except (SSLError, socket.error) as e:
                raise NetworkError(uri=server, error=str(e))
            except (OverflowError, TypeError) as e:
                raise XMLRPCMarshallError(error=str(e))
        raise NetworkError(
            uri=server,
            error=_("Exceeded number of tries to forward a request."))


class xmlclient(RPCClient):
    session_path = '/ipa/session/xml'
    server_proxy_class = ServerProxy
    protocol = 'xml'
    env_rpc_uri_key = 'xmlrpc_uri'

    def _call_command(self, command, params):
        version = params[1].get('version', VERSION_WITHOUT_CAPABILITIES)
        params = xml_wrap(params, version)
        result = command(*params)
        return xml_unwrap(result)


class JSONServerProxy:
    def __init__(self, uri, transport, encoding, verbose, allow_none):
        split_uri = urllib.parse.urlsplit(uri)
        if split_uri.scheme not in ("http", "https"):
            raise IOError("unsupported XML-RPC protocol")
        self.__host = split_uri.netloc
        self.__handler = split_uri.path
        self.__transport = transport

        assert encoding == 'UTF-8'
        assert allow_none
        self.__verbose = verbose

        # FIXME: Some of our code requires ServerProxy internals.
        # But, xmlrpc.client.ServerProxy's _ServerProxy__transport can be accessed
        # by calling serverproxy('transport')
        self._ServerProxy__transport = transport

    def __request(self, name, args):
        print_json = self.__verbose >= 2
        payload = {'method': unicode(name), 'params': args, 'id': 0}
        version = args[1].get('version', VERSION_WITHOUT_CAPABILITIES)
        payload = json_encode_binary(
            payload, version, pretty_print=print_json)

        if print_json:
            logger.info(
                'Request: %s',
                payload
            )

        response = self.__transport.request(
            self.__host,
            self.__handler,
            payload.encode('utf-8'),
            verbose=self.__verbose >= 3,
        )

        if print_json:
            logger.info(
                'Response: %s',
                json.dumps(json.loads(response), sort_keys=True, indent=4)
            )

        try:
            response = json_decode_binary(response)
        except ValueError as e:
            raise JSONError(error=str(e))

        error = response.get('error')
        if error:
            try:
                error_class = errors_by_code[error['code']]
            except KeyError:
                raise UnknownError(
                    code=error.get('code'),
                    error=error.get('message'),
                    server=self.__host,
                )
            else:
                kw = error.get('data', {})
                kw['message'] = error['message']
                raise error_class(**kw)

        return response['result']

    def __getattr__(self, name):
        def _call(*args):
            return self.__request(name, args)
        return _call


class jsonclient(RPCClient):
    session_path = '/ipa/session/json'
    server_proxy_class = JSONServerProxy
    protocol = 'json'
    env_rpc_uri_key = 'jsonrpc_uri'



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2009 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `ipalib.output` module.
"""

from ipatests.util import raises, ClassChecker
from ipalib import output
from ipalib.frontend import Command
from ipapython.version import API_VERSION

import pytest

pytestmark = pytest.mark.tier0

class test_Output(ClassChecker):
    """
    Test the `ipalib.output.Output` class.
    """

    _cls = output.Output

    def test_init(self):
        """
        Test the `ipalib.output.Output.__init__` method.
        """
        o = self.cls('result')
        assert o.name == 'result'
        assert o.type is None
        assert o.doc is None

    def test_repr(self):
        """
        Test the `ipalib.output.Output.__repr__` method.
        """
        o = self.cls('aye')
        assert repr(o) == "Output('aye')"
        o = self.cls('aye', type=int, doc='An A, aye?')
        assert repr(o) == (
            "Output('aye', type=[<type 'int'>], doc='An A, aye?')"
        )

        class Entry(self.cls):
            pass
        o = Entry('aye')
        assert repr(o) == "Entry('aye')"
        o = Entry('aye', type=int, doc='An A, aye?')
        assert repr(o) == (
            "Entry('aye', type=[<type 'int'>], doc='An A, aye?')"
        )


class test_ListOfEntries(ClassChecker):
    """
    Test the `ipalib.output.ListOfEntries` class.
    """

    _cls = output.ListOfEntries

    def test_validate(self):
        """
        Test the `ipalib.output.ListOfEntries.validate` method.
        """
        api = 'the api instance'
        class example(Command):
            pass
        cmd = example(api)
        inst = self.cls('stuff')

        okay = dict(foo='bar')
        nope = ('aye', 'bee')

        e = raises(TypeError, inst.validate,
                   cmd, [okay, okay, nope], API_VERSION)
        assert str(e) == output.emsg % (
            'example', 'ListOfEntries', 'stuff', 2, dict, tuple, nope
        )

        e = raises(TypeError, inst.validate,
                   cmd, [nope, okay, nope], API_VERSION)
        assert str(e) == output.emsg % (
            'example', 'ListOfEntries', 'stuff', 0, dict, tuple, nope
        )
